# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu


from geospacelab.datahub import FacilityModel


class GraceFacility(FacilityModel):
    def __new__(cls, str_in, **kwargs):
        obj = super().__new__(cls, str_in, **kwargs)
        return obj


grace_facility = GraceFacility('Grace')
grace_facility.url = 'https://earth.esa.int/eogateway/missions/grace'
grace_facility.category = 'GRACE online database'
grace_facility.Notes = ''



# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"

from geospacelab.datahub import DatabaseModel


class NCEIDatabase(DatabaseModel):
    def __new__(cls, str_in, **kwargs):
        obj = super().__new__(cls, str_in, **kwargs)
        return obj


ncei_database = NCEIDatabase('NCEI')
ncei_database.url = 'https://www.ncei.noaa.gov/data/'
ncei_database.category = 'online database'


import numpy as np
from scipy.signal import argrelextrema
import datetime
from scipy.interpolate import griddata, interp1d
from scipy.signal import butter, lfilter, freqz
import scipy.signal as sig

from geospacelab.datahub import DatasetUser
from geospacelab.toolbox.utilities import pydatetime as dttool
import geospacelab.toolbox.utilities.numpymath as npmath
from geospacelab.cs import GEOCSpherical


class LEOToolbox(DatasetUser):

    def __init__(self, dt_fr=None, dt_to=None, visual='on'):
        super().__init__(dt_fr, dt_to, visual=visual)

        self.add_variable(var_name='SC_GEO_LAT')
        self.add_variable(var_name='SC_GEO_LON')
        self.add_variable(var_name='SC_GEO_ALT')
        self.add_variable(var_name='SC_GEO_LST')
        self.add_variable(var_name='SC_DATETIME')
        self.add_variable(var_name='SC_AACGM_LAT')
        self.add_variable(var_name='SC_AACGM_LON')
        self.add_variable(var_name='SC_AACGM_MLT')
        self.add_variable(var_name='SC_APEX_LAT')
        self.add_variable(var_name='SC_APEX_LON')
        self.add_variable(var_name='SC_APEX_MLT')
        self.ascending_nodes = {}
        self.descending_nodes = {}
        self.northern_nodes = {}
        self.southern_nodes = {}
        self.sector_cs = 'GEO'
        self.sectors = {}

    def search_orbit_nodes(self, data_interval=1, t_res=None):
        def check_nodes(nodes):
            dts_nodes = nodes['DATETIME']
            sectimes_nodes, _ = dttool.convert_datetime_to_sectime(dts_nodes)
            diff_sectimes = np.diff(sectimes_nodes)
            if t_res is not None:
                sectime_res_1 = t_res
            else:
                sectime_res_1 = np.median(np.diff(sectimes_nodes))
            inds_abnormal = np.where(diff_sectimes<sectime_res_1*0.6)[0]
            if list(inds_abnormal):
                iiii = np.array(inds_abnormal) + 1
                nodes['INDEX'] = np.delete(nodes['INDEX'], iiii)
                nodes['GEO_LON'] = np.delete(nodes['GEO_LON'], iiii)
                nodes['GEO_LST'] = np.delete(nodes['GEO_LST'], iiii)
                nodes['DATETIME'] = np.delete(nodes['DATETIME'], iiii)


        glat = self['SC_GEO_LAT'].value.flatten()
        glat_ = glat[0::data_interval]
        glon = self['SC_GEO_LON'].value.flatten()
        lst = self['SC_GEO_LST'].value.flatten()
        dts = self['SC_DATETIME'].value.flatten()
        ind_1 = argrelextrema(np.abs(glat_), np.less, )[0]
        ind_1 = ind_1 * data_interval
        ind_2 = argrelextrema(glat_, np.greater, )[0]
        ind_2 = ind_2 * data_interval
        ind_3 = argrelextrema(glat_, np.less)[0]
        ind_3 = ind_3 * data_interval
        inds_asc = []
        inds_dsc = []
        for ind_N in ind_2:
            if ind_N < ind_1[0]:
                inds_dsc.append(ind_1[0])
            elif ind_N > ind_1[-1]:
                inds_asc.append(ind_1[-1])
            else:
                ind_between = np.where(ind_1 > ind_N)[0]
                inds_asc.append(ind_1[ind_between[0]-1])
                inds_dsc.append(ind_1[ind_between[0]])
        self.ascending_nodes['INDEX'] = inds_asc
        self.ascending_nodes['GEO_LON'] = glon[inds_asc]
        self.ascending_nodes['GEO_LST'] = lst[inds_asc]
        self.ascending_nodes['DATETIME'] = dts[inds_asc]
        check_nodes(self.ascending_nodes)

        self.descending_nodes['INDEX']= inds_dsc
        self.descending_nodes['GEO_LON'] = glon[inds_dsc]
        self.descending_nodes['GEO_LST'] = lst[inds_dsc]
        self.descending_nodes['DATETIME'] = dts[inds_dsc]
        check_nodes(self.descending_nodes)

        self.northern_nodes['INDEX'] = ind_2
        self.northern_nodes['GEO_LON'] = glon[ind_2]
        self.northern_nodes['GEO_LST'] = lst[ind_2]
        self.northern_nodes['DATETIME'] = dts[ind_2]
        check_nodes(self.northern_nodes)

        self.southern_nodes['INDEX'] = ind_3
        self.southern_nodes['GEO_LON'] = glon[ind_3]
        self.southern_nodes['GEO_LST'] = lst[ind_3]
        self.southern_nodes['DATETIME'] = dts[ind_3]
        check_nodes(self.southern_nodes)

    def group_by_sector(self, sector_name, boundary_lat, sector_cs='GEO'):
        self.sector_cs = sector_cs
        if not dict(self.ascending_nodes):
            self.search_orbit_nodes()

        glat = self['SC_GEO_LAT'].value.flatten()
        glon = self['SC_GEO_LON'].value.flatten()
        lst = self['SC_GEO_LST'].value.flatten()
        
        if self.sector_cs == 'AACGM':
            mlat = self['SC_AACGM_LAT'].value.flatten()
            mlon = self['SC_AACGM_LON'].value.flatten()
            mlt = self['SC_AACGM_MLT'].value.flatten()
        elif self.sector_cs == 'APEX':
            mlat = self['SC_APEX_LAT'].value.flatten()
            mlon = self['SC_APEX_LON'].value.flatten()
            mlt = self['SC_APEX_MLT'].value.flatten()
        
        dts = self['SC_DATETIME'].value.flatten()
        
        inds_asc = self.ascending_nodes['INDEX']
        inds_dsc = self.descending_nodes['INDEX']

        if self.sector_cs == 'GEO':
            lat = glat
            lt = lst
        elif self.sector_cs == 'AACGM':
            lat = mlat
            lt = mlt
        elif self.sector_cs == 'APEX':
            lat = mlat
            lt = mlt
        else:
            raise NotImplementedError

        # inds_center_N = argrelextrema(lat, np.greater)[0]
        inds_center_N = self.northern_nodes['INDEX']
        iii = np.where(lat[inds_center_N] > (np.max(lat[inds_center_N]))/1.5)[0]
        inds_center_N = inds_center_N[iii]
        # inds_center_N = inds_center_N[np.where(lat[inds_center_N] > boundary_lat)[0]]

        # inds_center_S = argrelextrema(lat, np.less)[0]
        inds_center_S = self.southern_nodes['INDEX']
        iii = np.where(lat[inds_center_S] < (np.min(lat[inds_center_S]))/1.5)[0]
        inds_center_S = inds_center_S[iii]
        # inds_center_S = inds_center_S[np.where(lat[inds_center_S] < -np.abs(boundary_lat))[0]]

        # Sector centered at the northern pole from the ascending node towards the descending node.
        if sector_name == 'N':
            sector = np.zeros((dts.size,))
            dts_c = np.zeros_like(dts)

            pseudo_lat = np.empty_like(sector)
            pseudo_lat[:] = np.nan
            for num, ind_rec in enumerate(inds_center_N):

                ind_tmp = np.where(inds_center_S < ind_rec)[0]
                if list(ind_tmp):
                    ind_rec_S1 = inds_center_S[ind_tmp[-1]]
                else:
                    ind_rec_S1 = 0
                ind_tmp = np.where(inds_center_S > ind_rec)[0]
                if list(ind_tmp):
                    ind_rec_S2 = inds_center_S[ind_tmp[0]]
                else:
                    ind_rec_S2 = len(lat) - 1
                inds_seg = np.array(range(ind_rec_S1, ind_rec_S2))

                dt_c = dts[ind_rec]
                lat_seg = lat[inds_seg]
                dts_seg = dts[inds_seg]
                delta_t_seg = np.array([(dt1 - dt_c).total_seconds() / 60. for dt1 in dts_seg])
                inds_seg_seg = np.where((lat_seg > boundary_lat) & (np.abs(delta_t_seg) < 60.))[0]

                inds_sector = inds_seg[inds_seg_seg]
                sector[inds_sector] = num + 1
                dts_c[inds_sector] = dt_c


                lat_seg = np.where(dts_seg < dt_c, lat_seg, 180. - lat_seg)
                pseudo_lat[inds_sector] = lat_seg[inds_seg_seg]

            self.add_variable(var_name='SECTOR_N', value=sector.reshape((sector.size, 1)))
            self.add_variable(var_name='SECTOR_N_DATETIME', value=dts_c.reshape((sector.size, 1)))
            self.add_variable(var_name='SECTOR_N_PSEUDO_LAT', value=pseudo_lat.reshape((sector.size, 1)))
            self.sectors['N'] = {
                'BOUNDARY_LAT': boundary_lat,
                'PSEUDO_LAT_RANGE': [boundary_lat, 180. - boundary_lat],
                'VARIABLE_NAMES': ['SECTOR_N', 'SECTOR_N_DATETIME', 'SECTOR_N_PSEUDO_LAT']
            }
        
        # Sector centered at the southern pole from the descending node towards the ascending node.
        elif sector_name == 'S':
            sector = np.zeros((dts.size,))
            dts_c = np.zeros_like(dts)

            pseudo_lat = np.empty_like(sector)
            pseudo_lat[:] = np.nan
            for num, ind_rec in enumerate(inds_center_S):

                ind_tmp = np.where(inds_center_N < ind_rec)[0]
                if list(ind_tmp):
                    ind_rec_N1 = inds_center_N[ind_tmp[-1]]
                else:
                    ind_rec_N1 = 0
                ind_tmp = np.where(inds_center_N > ind_rec)[0]
                if list(ind_tmp):
                    ind_rec_N2 = inds_center_N[ind_tmp[0]]
                else:
                    ind_rec_N2 = len(lat) - 1
                inds_seg = np.array(range(ind_rec_N1, ind_rec_N2))

                dt_c = dts[ind_rec]
                lat_seg = lat[inds_seg]
                dts_seg = dts[inds_seg]
                delta_t_seg = np.array([(dt1 - dt_c).total_seconds() / 60. for dt1 in dts_seg])
                inds_seg_seg = np.where((lat_seg < -np.abs(boundary_lat)) & (np.abs(delta_t_seg) < 60.))[0]

                inds_sector = inds_seg[inds_seg_seg]
                sector[inds_sector] = num + 1
                dts_c[inds_sector] = dt_c

                lat_seg = np.where(dts_seg < dt_c, 180. - lat_seg, 360. + lat_seg)
                pseudo_lat[inds_sector] = lat_seg[inds_seg_seg]

            self.add_variable(var_name='SECTOR_S', value=sector.reshape((sector.size, 1)))
            self.add_variable(var_name='SECTOR_S_DATETIME', value=dts_c.reshape((sector.size, 1)))
            self.add_variable(var_name='SECTOR_S_PSEUDO_LAT', value=pseudo_lat.reshape((sector.size, 1)))
            self.sectors['S'] = {
                'BOUNDARY_LAT': boundary_lat,
                'PSEUDO_LAT_RANGE': [180. + np.abs(boundary_lat), 360. - np.abs(boundary_lat)],
                'VARIABLE_NAMES': ['SECTOR_S', 'SECTOR_S_DATETIME', 'SECTOR_S_PSEUDO_LAT']
            }

        # Sector centered at the asending node from south towards north.
        elif sector_name == 'ASC':
            sector = np.zeros((dts.size,))
            dts_c = np.zeros_like(dts)

            pseudo_lat = np.empty_like(sector)
            pseudo_lat[:] = np.nan
            for num, ind_rec in enumerate(inds_asc):

                ind_tmp = np.where(inds_center_S < ind_rec)[0]
                if list(ind_tmp):
                    ind_rec_S1 = inds_center_S[ind_tmp[-1]]
                else:
                    ind_rec_S1 = 0
                ind_tmp = np.where(inds_center_N > ind_rec)[0]
                if list(ind_tmp):
                    ind_rec_N1 = inds_center_N[ind_tmp[0]]
                else:
                    ind_rec_N1 = len(lat) - 1 
                inds_seg = np.array(range(ind_rec_S1, ind_rec_N1))

                dt_c = dts[ind_rec]
                lat_seg = lat[inds_seg]
                dts_seg = dts[inds_seg]
                delta_t_seg = np.array([(dt1 - dt_c).total_seconds() / 60. for dt1 in dts_seg])
                inds_seg_seg = np.where((np.abs(lat_seg) < np.abs(boundary_lat)) & (np.abs(delta_t_seg) < 60.))[0]

                inds_sector = inds_seg[inds_seg_seg]
                sector[inds_sector] = num + 1
                dts_c[inds_sector] = dt_c

                lat_seg = np.where(lat_seg < 0, lat_seg, lat_seg)
                pseudo_lat[inds_sector] = lat_seg[inds_seg_seg]

            self.add_variable(var_name='SECTOR_ASC', value=sector.reshape((sector.size, 1)))
            self.add_variable(var_name='SECTOR_ASC_DATETIME', value=dts_c.reshape((sector.size, 1)))
            self.add_variable(var_name='SECTOR_ASC_PSEUDO_LAT', value=pseudo_lat.reshape((sector.size, 1)))
            self.sectors['ASC'] = {
                'BOUNDARY_LAT': boundary_lat,
                'PSEUDO_LAT_RANGE': [-np.abs(boundary_lat), np.abs(boundary_lat)],
                'VARIABLE_NAMES': ['SECTOR_ASC', 'SECTOR_ASC_DATETIME', 'SECTOR_ASC_PSEUDO_LAT']
            }
            
            # Sector centered at the asending node from south towards north.
        elif sector_name == 'DSC':
            sector = np.zeros((dts.size,))
            dts_c = np.zeros_like(dts)

            pseudo_lat = np.empty_like(sector)
            pseudo_lat[:] = np.nan
            for num, ind_rec in enumerate(inds_dsc):

                ind_tmp = np.where(inds_center_N < ind_rec)[0]
                if list(ind_tmp):
                    ind_rec_N1 = inds_center_N[ind_tmp[-1]]
                else:
                    ind_rec_N1 = 0
                ind_tmp = np.where(inds_center_S > ind_rec)[0]
                if list(ind_tmp):
                    ind_rec_S1 = inds_center_S[ind_tmp[0]]
                else:
                    ind_rec_S1 = len(lat) - 1
                inds_seg = np.array(range(ind_rec_N1, ind_rec_S1))

                dt_c = dts[ind_rec]
                lat_seg = lat[inds_seg]
                dts_seg = dts[inds_seg]
                delta_t_seg = np.array([(dt1 - dt_c).total_seconds() / 60. for dt1 in dts_seg])
                inds_seg_seg = np.where((np.abs(lat_seg) < np.abs(boundary_lat)) & (np.abs(delta_t_seg) < 60.))[0]

                inds_sector = inds_seg[inds_seg_seg]
                sector[inds_sector] = num + 1
                dts_c[inds_sector] = dt_c

                lat_seg = np.where(lat_seg > 0, 180. - lat_seg, 180. - lat_seg)
                pseudo_lat[inds_sector] = lat_seg[inds_seg_seg]

            self.add_variable(var_name='SECTOR_DSC', value=sector.reshape((sector.size, 1)))
            self.add_variable(var_name='SECTOR_DSC_DATETIME', value=dts_c.reshape((sector.size, 1)))
            self.add_variable(var_name='SECTOR_DSC_PSEUDO_LAT', value=pseudo_lat.reshape((sector.size, 1)))
            self.sectors['DSC'] = {
                'BOUNDARY_LAT': boundary_lat,
                'PSEUDO_LAT_RANGE': [180. - np.abs(boundary_lat), 180. + np.abs(boundary_lat)],
                'VARIABLE_NAMES': ['SECTOR_DSC', 'SECTOR_DSC_DATETIME', 'SECTOR_DSC_PSEUDO_LAT']
            }
        else:
            raise NotImplementedError
                
    def griddata_by_sector(
            self, sector_name=None, variable_names=None, x_grid_res=20*60, y_grid_res=0.5, along_track_interp=True,
            x_data_res=None, y_data_res=None, along_track_binning=False, 
            ):
        self.visual = 'on'
        dts_c = self['_'.join(('SECTOR', sector_name, 'DATETIME'))].value.flatten()
        dts = self['SC_DATETIME'].value.flatten()
        sector = self['SECTOR_' + sector_name].value.flatten()
        lat = self['_'.join(('SECTOR', sector_name, 'PSEUDO_LAT'))].value.flatten()
        boundary_lat = self.sectors[sector_name]['BOUNDARY_LAT']
        lat_range = self.sectors[sector_name]['PSEUDO_LAT_RANGE']

        if along_track_interp:
            x_data = dts[sector > 0]
        else:
            x_data = dts_c[sector > 0]
        
        x_data_1 = dts_c[sector>0]
        y_data = lat[sector > 0]
        is_finite = np.isfinite(y_data)
        dt0 = dttool.get_start_of_the_day(self.dt_fr)
        sectime = np.array([(dt - dt0).total_seconds() for dt in x_data])
        x = sectime
        y = y_data
        
        x_1 = np.array([(dt - dt0).total_seconds() for dt in x_data_1]) 
        # max_x = np.ceil(np.max(x) / self.xgrid_res) * self.xgrid_res
        # min_x = np.floor(np.min(x) / self.xgrid_res) * self.xgrid_res

        x_unique = np.unique(dts_c[sector > 0])
        x_unique = [(t - dt0).total_seconds() for t in x_unique]
        if x_data_res is None:
            x_data_res = np.median(np.diff(x_unique))

        if y_data_res is None:
            y_data_res = np.median(np.diff(y_data[y_data < np.abs(np.diff(lat_range))]/2))
            if y_grid_res > y_data_res:
                y_data_res = y_grid_res

        ny = int(np.ceil(np.diff(lat_range) / y_grid_res) + 1)
        if x_grid_res is None:
            grid_x, grid_y = np.meshgrid(x_unique, np.linspace(lat_range[0], lat_range[1], ny))
            x_interp = False
            x_grid_res=x_data_res
        else:
            min_x = np.floor((self.dt_fr - dt0).total_seconds() / x_grid_res) * x_grid_res
            max_x = np.ceil((self.dt_to - dt0).total_seconds() / x_grid_res) * x_grid_res
            # total_seconds = (dts[-1] - dt0).total_seconds()
            # min_x = 0
            # max_x = total_seconds
            # nx = int(np.ceil((max_x - min_x) / x_grid_res))
            # grid_x, grid_y = np.meshgrid(np.linspace(min_x, max_x, nx),
            #                            np.linspace(lat_range[0], lat_range[1], ny))
            grid_x, grid_y = np.meshgrid(
                np.arange(min_x, max_x, x_grid_res),
                np.linspace(lat_range[0], lat_range[1], ny)
            )
            x_interp = True

        grid_lat = griddata((x_1, y), y, (grid_x, grid_y), method='nearest')

        if self.sector_cs == 'AACGM':
            mask_y = np.abs(grid_y - grid_lat) > y_data_res * 2
            which_lat = 'AACGM_LAT'
        if self.sector_cs == 'APEX':
            mask_y = np.abs(grid_y - grid_lat) > y_data_res * 2
            which_lat = 'APEX_LAT' 
        else:
            mask_y = np.abs(grid_y - grid_lat) > y_data_res * 1.2
            which_lat = 'GEO_LAT'

        grid_sectime = griddata((x_1, y), x_1, (grid_x, grid_y), method='nearest')
        mask_x = np.abs(grid_x - grid_sectime) > x_data_res * 1.2

        # remove gaps
        i_mask_x = np.where(mask_x)
        rec_bounds_ind = {}
        if list(i_mask_x[0]):
            for ii, jj in zip(*i_mask_x):
                xx = grid_x[ii, jj]
                x_ii = grid_x[ii, :].flatten()
                if xx in rec_bounds_ind.keys():
                    bound_x_1 = rec_bounds_ind[xx][0]
                    bound_x_2 = rec_bounds_ind[xx][1]
                else:
                    iii = np.where(x_unique<xx)[0]
                    if not list(iii):
                        bound_x_1 = x_ii[0] - x_data_res
                    else:
                        bound_x_1 = x_unique[iii[-1]] - x_grid_res
                    iii = np.where(x_unique > xx)[0]
                    if not list(iii):
                        bound_x_2 = x_ii[-1] + x_grid_res
                    else:
                        bound_x_2 = x_unique[iii[0]] + x_data_res
                    rec_bounds_ind[xx] = [bound_x_1, bound_x_2]
                mask_x[ii, (x_ii >= bound_x_1) & (x_ii <= bound_x_2)] = True
            pass

        method = 'linear'
        for var_name in variable_names:
            # print(var_name)
            vrb = self[var_name].value.flatten()[sector>0]
            if along_track_binning:
                sector_1 = sector[sector>0]
                n_tracks = int(np.max(sector))
                n_lats = ny
                xx_1 = np.empty((n_tracks, n_lats))
                zz_1 = np.empty((n_tracks, n_lats))
                yy_1 = grid_y[:, 0].flatten()
                grid_z = np.ones_like(grid_x)*np.nan
                for ii in np.arange(1, n_tracks+1):
                    xd = x[sector_1 == ii]
                    yd = y[sector_1 == ii]
                    zd = vrb[sector_1 == ii]
                    if (not list(xd)) or (not list(yd)):
                        continue
                        
                    f = interp1d(yd, xd, bounds_error=False, fill_value= 'extrapolate')
                    x_i = f(yy_1)
                    xx_1[ii-1, :] = x_i

                    if 'LON' in var_name:
                        z_i = npmath.interp_period_data(yd, zd, yy_1,  period=360., method='linear', bounds_error=False)
                    elif 'LST' in var_name:
                        z_i = npmath.interp_period_data(yd, zd, yy_1,  period=24., method='linear', bounds_error=False)
                    else:
                        # f = interp1d(yd, zd, bounds_error=False, fill_value='extrapolate')
                        # z_i = f(yy_1)
                        z_i = np.ones_like(yy_1) * np.nan       
                        for j, yyy in enumerate(yy_1):
                            inds_y = np.where((yd >= yyy-y_grid_res/2) & (yd < yyy+y_grid_res/2))[0]
                            if list(inds_y):
                                z_i[j] = np.nanmean(zd[inds_y])
                    zz_1[ii-1, :] = z_i 
                for ii in range(grid_x.shape[0]):
                    xd = xx_1[:, ii].flatten()
                    zd = zz_1[:, ii].flatten()
                    if x_interp:
                        if 'LON' in var_name:
                            z_i = npmath.interp_period_data(xd, zd, grid_x[0],  period=360., method='linear', bounds_error=False)
                        elif 'LST' in var_name:
                            z_i = npmath.interp_period_data(xd, zd, grid_x[0],  period=24., method='linear', bounds_error=False)
                        else:
                            f = interp1d(xd, zd, bounds_error=False, fill_value='extrapolate')
                            z_i = f(grid_x[0])
                        grid_z[ii, :] = z_i
                    else:
                        grid_z[ii, :] = zd 
            elif along_track_interp:
                sector_1 = sector[sector>0]
                n_tracks = int(np.max(sector))
                n_lats = ny
                xx_1 = np.empty((n_tracks, n_lats))
                zz_1 = np.empty((n_tracks, n_lats))
                yy_1 = grid_y[:, 0].flatten()
                grid_z = np.ones_like(grid_x) * np.nan
                for ii in np.arange(1, n_tracks+1):
                    xd = x[sector_1 == ii]
                    yd = y[sector_1 == ii]
                    zd = vrb[sector_1 == ii]
                    
                    if (not list(xd)) or (not list(yd)):
                        continue

                    f = interp1d(yd, xd, bounds_error=False, fill_value= 'extrapolate')
                    x_i = f(yy_1)
                    xx_1[ii-1, :] = x_i

                    if 'LON' in var_name:
                        z_i = npmath.interp_period_data(yd, zd, yy_1,  period=360., method='linear', bounds_error=False)
                    elif 'LST' in var_name:
                        z_i = npmath.interp_period_data(yd, zd, yy_1,  period=24., method='linear', bounds_error=False)
                    else:
                        f = interp1d(yd, zd, bounds_error=False, fill_value='extrapolate')
                        z_i = f(yy_1)
                    zz_1[ii-1, :] = z_i 
                for ii in range(grid_x.shape[0]):
                    xd = xx_1[:, ii].flatten()
                    zd = zz_1[:, ii].flatten()
                    if x_interp:
                        if 'LON' in var_name:
                            z_i = npmath.interp_period_data(xd, zd, grid_x[0],  period=360., method='linear', bounds_error=False)
                        elif 'LST' in var_name:
                            z_i = npmath.interp_period_data(xd, zd, grid_x[0],  period=24., method='linear', bounds_error=False)
                        else:
                            f = interp1d(xd, zd, bounds_error=False, fill_value='extrapolate')
                            z_i = f(grid_x[0])
                        grid_z[ii, :] = z_i
                    else:
                        grid_z[ii, :] = zd
            else:
                grid_z = griddata((x, y), vrb, (grid_x, grid_y), method=method)
            grid_z[mask_y] = np.nan
            grid_z[mask_x] = np.nan
            var_name_in = '_'.join(('SECTOR', sector_name, 'GRID', var_name))
            self.sectors[sector_name]['VARIABLE_NAMES'].append(var_name_in) 
            
            # self.add_variable(var_name=var_name_in, value=grid_z.T)
            self[var_name_in] = self[var_name].clone(omit_attrs='visual')
            self[var_name_in].visual = 'new'
            var = self[var_name_in]
            var.value = grid_z.T
            var.set_depend(0, {'UT': '_'.join(('SECTOR', sector_name, 'GRID_DATETIME'))})
            var.set_depend(1, {which_lat: '_'.join(('SECTOR', sector_name, 'GRID_LAT'))})
            var.visual.axis[0].data = '@d.' + '_'.join(('SECTOR', sector_name, 'GRID_DATETIME'))
            var.visual.axis[1].data = '@d.' + '_'.join(('SECTOR', sector_name, 'GRID_LAT'))
            var.visual.axis[1].label = 'GLAT' if self.sector_cs == 'GEO' else 'MLAT'
            var.visual.axis[1].unit = r'$^\circ$'
            var.visual.axis[1].lim = lat_range
            var.visual.axis[2].data = '@v.value'
            var.visual.axis[2].data_scale = self[var_name].visual.axis[1].data_scale
            var.visual.axis[2].label = '@v.label'
            var.visual.axis[2].unit = '@v.unit'
            var.visual.plot_config.style = '2P'
            var.visual.plot_config.pcolormesh.update(cmap='jet')

        grid_x_name = '_'.join(('SECTOR', sector_name, 'GRID', 'X'))
        grid_x = grid_x.T
        self.add_variable(var_name=grid_x_name, value=grid_x)
        grid_y_name = '_'.join(('SECTOR', sector_name, 'GRID', 'Y'))
        grid_y = grid_y.T
        self.add_variable(var_name=grid_y_name, value=grid_y)
        var_name = '_'.join(('SECTOR', sector_name, 'GRID_DATETIME'))
        grid_dts = np.array([dt0 + datetime.timedelta(seconds=sec) for sec in grid_x[:, 0]])
        self.add_variable(var_name=var_name, value=grid_dts[:, np.newaxis])
        var_name = '_'.join(('SECTOR', sector_name, 'GRID_LAT'))
        self.add_variable(var_name=var_name, value=grid_y[0, :][np.newaxis, :])

    def filter_by_time(self, variable_names=None, time_window=40., time_res=None):


        return
    
    def add_GEO_LST(self):
        lons = self['SC_GEO_LON'].flatten()
        uts = self['SC_DATETIME'].flatten()
        lsts = [ut + datetime.timedelta(seconds=int(lon / 15. * 3600)) for ut, lon in zip(uts, lons)]
        lsts = [lst.hour + lst.minute / 60. + lst.second / 3600. for lst in lsts]
        var = self.add_variable(var_name='SC_GEO_LST')
        var.value = np.array(lsts)[:, np.newaxis]
        var.label = 'LST'
        var.unit = 'h'
        var.depends = self['SC_GEO_LON'].depends
        return var
    
    @staticmethod
    def format_pseudo_lat_label(ax, sector_name, is_integer=True, y_tick_res=15.):
        if is_integer:
            lat_format = '{:4.0f}'
        else:
            lat_format = '{:5.1f}'
    
        y_lim = ax.get_ylim()
        y_max = np.max(y_lim)
        y_min = np.min(y_lim)
        
        yticks_ref = np.arange(-360, 720 + y_tick_res, y_tick_res)
        iii = np.where((yticks_ref>=y_min) & (yticks_ref<=y_max))[0]
        # yticks = ax.get_yticks()
        yticks = yticks_ref[iii]
        ylabels = []
        n1 = 0
        n2 = len(yticks) - 1
        n3 = 0
        for ind, pseudo_lat in enumerate(yticks):
            if pseudo_lat < y_min:
                n1 = n1 + 1
                ylabels.append('')
                continue
            if pseudo_lat > y_max:
                n2 = n2 - 1
                ylabels.append('')
                continue
            if sector_name == 'N':
                lat = pseudo_lat if pseudo_lat <= 90. else 180. - pseudo_lat
                if lat == 90.:
                    n3 = ind

            elif sector_name == 'S':
                lat = 180. - pseudo_lat if pseudo_lat <= 270. else pseudo_lat - 360.
                if lat == -90.:
                    n3 = ind

            elif sector_name == 'ASC':
                lat = pseudo_lat
                if lat == 0.:
                    n3 = ind

            elif sector_name == 'DSC':
                lat = 180. - pseudo_lat
                if lat == 0.:
                    n3 = ind

            else:
                raise NotImplementedError
            if lat > 0:
                lat_format_1 = r'' + lat_format + r'$^\circ$N'
                lb = lat_format_1.format(lat)
            elif lat < 0:
                lat_format_1 = r'' + lat_format + r'$^\circ$S'
                lb = lat_format_1.format(np.abs(lat))
            else:
                lat_format_1 = r'' + '{:4.0f}' + r'$^\circ$'
                lb = lat_format_1.format(lat) 
            ylabels.append(lb)

        if sector_name == 'N':
            ylabels[n1] = 'ASC:' + ylabels[n1]
            ylabels[n2] = 'DSC:' + ylabels[n2]
            ylabels[n3] = 'Pole->' + ylabels[n3]
        elif sector_name == 'S':
            ylabels[n1] = 'DSC:' + ylabels[n1]
            ylabels[n2] = 'ASC:' + ylabels[n2]
            ylabels[n3] = 'Pole->' + ylabels[n3]
        elif sector_name == 'ASC':
            ylabels[n1] = '' + ylabels[n1]
            ylabels[n2] = '' + ylabels[n2]
            ylabels[n3] = 'ASC->' + ylabels[n3]
        elif sector_name == 'DSC':
            ylabels[n1] = '' + ylabels[n1]
            ylabels[n2] = '' + ylabels[n2]
            ylabels[n3] = 'DSC->' + ylabels[n3]
        else:
            raise NotImplementedError
        ax.set_yticks(yticks)
        ax.set_yticklabels(ylabels)
        ax.set_ylim(y_lim)

        return

    def smooth_along_track(self, time_window=450, time_res=None, variable_names=None, method='savgol', residuals='on'):
        self.visual = 'on'
        dts = self['SC_DATETIME'].value.flatten()
        dt0 = dttool.get_start_of_the_day(self.dt_fr)
        sectime = np.array([(dt - dt0).total_seconds() for dt in dts])
        if time_res is None:
            diff_secs = np.diff(sectime)
            time_res = np.median(diff_secs)

        N = int(np.fix(time_window/time_res))

        for var_name in variable_names:
            vrb = self[var_name].value.flatten()
            vrb_avg = self.smooth_savgol(dts, vrb, N, time_res)
            var_name_in = var_name + '_AVG'
            self[var_name_in] = self[var_name].clone()
            self[var_name_in].value = vrb_avg[:, np.newaxis]
            self[var_name_in].label = 'smoothed ' + self[var_name_in].label
            if residuals:
                var_name_in = var_name + '_RSD'
                self[var_name_in] = self[var_name].clone()
                self[var_name_in].value = self[var_name].value - self[var_name + '_AVG'].value
                self[var_name_in].label = 'RSD'
                self[var_name_in].visual.axis[1].lim = [-np.inf, np.inf]
                var_name_in = var_name + '_RSD_PERC'
                self[var_name_in] = self[var_name].clone()
                self[var_name_in].value = (self[var_name].value - self[var_name + '_AVG'].value) / \
                    self[var_name + '_AVG'].value
                self[var_name_in].label = "RSD pct."
                self[var_name_in].unit = '%'
                self[var_name_in].unit_label = None
                self[var_name_in].visual.axis[1].data_scale = 100.

    @staticmethod
    def smooth_savgol(dts, data, box_pts, t_res, poly_order=1):
        delta_sectime = 1
        sectime, _ = dttool.convert_datetime_to_sectime(dts)
        ind_nan = np.isnan(data)
        sectime_i = np.arange(0, sectime[-1], delta_sectime)
        data_i = np.interp(sectime_i, sectime[~ind_nan], data[~ind_nan])

        window_size = int(np.floor(t_res / delta_sectime * box_pts / 2) * 2 + 1)
        y = sig.savgol_filter(data_i, window_size, poly_order)
        data_new = np.interp(sectime, sectime_i, y)
        data_new[ind_nan] = np.nan
        return data_new
    
    def trajectory_local_unit_vector(self):
        lat_0 = self['SC_GEO_LAT'].flatten()
        lon_0 = self['SC_GEO_LON'].flatten()
        height_0 = self['SC_GEO_ALT'].flatten()
        
        cs_0 = GEOCSpherical(coords={'lat': lat_0, 'lon': lon_0, 'height': height_0})
        phi_0 = cs_0['phi']
        theta_0 = cs_0['theta']
        cs_0 = cs_0.to_cartesian()
        x_0 = cs_0['x']
        y_0 = cs_0['y']
        z_0 = cs_0['z']
        
        dx = np.concatenate(
            (
                [cs_0['x'][1] - cs_0['x'][0]], 
                cs_0['x'][2:] - cs_0['x'][0:-2],
                [cs_0['x'][-1] - cs_0['x'][-2]]
            ),
            axis=0
        )
        dy = np.concatenate(
            (
                [cs_0['y'][1] - cs_0['y'][0]], 
                cs_0['y'][2:] - cs_0['y'][0:-2],
                [cs_0['y'][-1] - cs_0['y'][-2]]
            ),
            axis=0
        )
        dz = np.concatenate(
            (
                [cs_0['z'][1] - cs_0['z'][0]], 
                cs_0['z'][2:] - cs_0['z'][0:-2],
                [cs_0['z'][-1] - cs_0['z'][-2]]
            ),
            axis=0
        )
        
        dv = np.array([dx.flatten(), dy.flatten(), dz.flatten()]).T
        
        v_new = np.empty_like(dv)
        
        for ind, (phi_c, theta_c) in enumerate(zip(phi_0, theta_0)):
            R_1 = np.array([
                [-np.sin(phi_c), -np.cos(phi_c), 0],
                [np.cos(phi_c), -np.sin(phi_c), 0],
                [0, 0, 1]
            ])

            R_2 = np.array([
                [1, 0, 0],
                [0, np.cos(theta_c), -np.sin(theta_c)],
                [0, np.sin(theta_c), np.cos(theta_c)]
            ])
            v_new[ind, :] = dv[ind, :].reshape((1, 3)) @ R_1 @ R_2
            
        
        norm = np.sqrt(v_new[:, 0]**2 + v_new[:, 1]**2 + v_new[:, 2]**2)
        
        v_unit = np.empty_like(v_new)
        v_unit[:, 0] = v_new[:, 1] / norm
        v_unit[:, 1] = v_new[:, 0] / norm
        v_unit[:, 2] = - v_new[:, 2] / norm
        
        return v_unit

         


# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"


import datetime
import numpy as np
import functools

import geospacelab.datahub as datahub
from geospacelab.datahub import DatabaseModel, FacilityModel
from geospacelab.datahub.sources.cdaweb import cdaweb_database
from geospacelab.config import prf
import geospacelab.toolbox.utilities.pydatetime as dttool
import geospacelab.toolbox.utilities.pybasic as basic
import geospacelab.toolbox.utilities.pylogging as mylog
from geospacelab.datahub.sources.cdaweb.omni.loader import Loader as default_Loader
from geospacelab.datahub.sources.cdaweb.omni.downloader import Downloader as default_Downloader
import geospacelab.datahub.sources.cdaweb.omni.variable_config as var_config


default_dataset_attrs = {
    'facility': 'OMNI',
    'omni_type': 'omni2',
    'omni_res': '1min',
    'data_file_type': 'hres-cdf',
    'data_file_ext': 'cdf',
    'data_root_dir': prf.datahub_data_root_dir / 'CDAWeb' / 'OMNI',
    'allow_load': False,
    'allow_download': True,
    'data_search_recursive': False,
    'label_fields': ['database', 'facility', 'omni_type', 'omni_res'],
    'time_clip': True,
}

default_variable_names = ['DATETIME', 'SC_ID_IMF', 'SC_ID_PLS', 'IMF_PTS', 'PLS_PTS', 'PCT_INTERP',
                          'Timeshift', 'Timeshift_RMS',
                          'B_x_GSE', 'B_y_GSE', 'B_z_GSE',
                          'B_x_GSM', 'B_y_GSM', 'B_z_GSM', 'B_T_GSM', 'B_TOTAL',
                          'v_sw', 'v_x', 'v_y', 'v_z',
                          'n_p', 'T', 'p_dyn', 'E', 'beta', 'Ma_A', 'Ma_MSP',
                          'BSN_x', 'BSN_y', 'BSN_z',
                          'AE', 'AL', 'AU', 'SYM_H', 'SYM_D', 'ASY_H', 'ASY_D']

# default_data_search_recursive = True

default_attrs_required = ['omni_type', 'omni_res']


def _validate_IMF_cs(func):
    @functools.wraps(func)
    def wrapper(ds_omni, **kwargs):
        kwargs.setdefault('cs', 'GSM')
        kwargs.setdefault('to_radian', False)
        cs = kwargs['cs']
        if cs == 'GSM':
            kwargs.setdefault('Bx', ds_omni['B_x_GSM'].flatten())
            kwargs.setdefault('By', ds_omni['B_y_GSM'].flatten())
            kwargs.setdefault('Bz', ds_omni['B_z_GSM'].flatten())
        elif cs == 'GSE':
            kwargs.setdefault('Bx', ds_omni['B_x_GSE'].flatten())
            kwargs.setdefault('By', ds_omni['B_y_GSE'].flatten())
            kwargs.setdefault('Bz', ds_omni['B_z_GSE'].flatten())
        else:
            raise NotImplementedError
        result = func(ds_omni, **kwargs)

        if kwargs['to_radian']:
            unit = 'degree'
            unit_label = r'$^\circ$'
        else:
            unit = 'radiance'
            unit_label = r''
        result.unit = unit
        result.unit_label = unit_label
        result.depends = ds_omni['B_x_GSE'].depends
        if ds_omni.visual == 'on':
            result.visual.plot_config.style = '1P'
            result.visual.axis[1].unit = '@v.unit_label'
        return result

    return wrapper


class Dataset(datahub.DatasetSourced):
    def __init__(self, **kwargs):
        kwargs = basic.dict_set_default(kwargs, **default_dataset_attrs)

        super().__init__(**kwargs)

        self.database = cdaweb_database
        self.facility = kwargs.pop('facility', '')
        self.omni_type = kwargs.pop('omni_type', 'omni2')
        self.omni_res = kwargs.pop('omni_res', '1min')
        self.data_file_type = kwargs.pop('data_file_type','')
        self.allow_download = kwargs.pop('allow_download', True)

        self.metadata = None

        allow_load = kwargs.pop('allow_load', False)

        # self.config(**kwargs)

        if self.loader is None:
            self.loader = default_Loader

        if self.downloader is None:
            self.downloader = default_Downloader

        self._set_default_variables(
            default_variable_names,
            configured_variables=var_config.configured_variables
        )

        self._validate_attrs()

        if allow_load:
            self.load_data()

    def _validate_attrs(self):
        for attr_name in default_attrs_required:
            attr = getattr(self, attr_name)
            if not str(attr):
                mylog.StreamLogger.warning("The parameter {} is required before loading data!".format(attr_name))

        if str(self.data_file_type):
            self.data_file_ext = self.data_file_type.split('-')[1]

    def label(self, **kwargs):
        label = super().label()
        return label

    def load_data(self, **kwargs):
        self.check_data_files(**kwargs)

        for file_path in self.data_file_paths:
            load_obj = self.loader(file_path, file_type=self.data_file_type)

            for var_name in self._variables.keys():
                self._variables[var_name].join(load_obj.variables[var_name])

            self.metadata = load_obj.metadata
            # self.select_beams(field_aligned=True)
        if self.time_clip:
            self.time_filter_by_range()

    def search_data_files(self, **kwargs):
        dt_fr = self.dt_fr
        dt_to = self.dt_to
        diff_months = dttool.get_diff_months(dt_fr, dt_to)
        dt0 = dttool.get_first_day_of_month(dt_fr)
        for i in range(diff_months + 1):
            thismonth = dttool.get_next_n_months(dt0, i)
            if self.omni_res in ['1min', '5min']:
                initial_file_dir = kwargs.pop('initial_file_dir', None)
                if initial_file_dir is None:
                    initial_file_dir = self.data_root_dir / \
                                       (self.omni_type.upper() + '_high_res_' + self.omni_res) / \
                                       '{:4d}'.format(thismonth.year)
                file_patterns = [
                    self.omni_res,
                    thismonth.strftime('%Y%m%d')
                ]
            else:
                raise NotImplementedError
            # remove empty str
            search_pattern = kwargs.pop('search_pattern', None)
            if search_pattern is None:
                file_patterns = [pattern for pattern in file_patterns if str(pattern)]

                search_pattern = '*' + '*'.join(file_patterns) + '*'

            done = super().search_data_files(
                initial_file_dir=initial_file_dir, search_pattern=search_pattern)

            # Validate file paths

            if not done and self.allow_download:
                done = self.download_data()
                if done:
                    done = super().search_data_files(
                        initial_file_dir=initial_file_dir, search_pattern=search_pattern)
                else:
                    print('Cannot find files from the online database!')

        return done

    def download_data(self):
        if self.data_file_type == 'hres-cdf':
            if self.omni_type.upper() == 'OMNI':
                new_omni = False
            elif self.omni_type.upper() == 'OMNI2':
                new_omni = True
            download_obj = self.downloader(dt_fr=self.dt_fr, dt_to=self.dt_to,
                                           res=self.omni_res, new_omni=new_omni,
                                           data_file_root_dir=self.data_root_dir)
        else:
            raise NotImplementedError
        return download_obj.done

    # _validate_IMF_cs = staticmethod(_validate_IMF_cs)

    @_validate_IMF_cs
    def add_IMF_CA(self, cs='GSM', to_radian=False,  **kwargs):
        """
        IMF clock angle in the IMF By-Bz (x-y) plane in the GSE/GSM coordinate system.
        :return:
        """
        Bz = kwargs['Bz']
        By = kwargs['By']
        Bt = np.sqrt(By**2+Bz**2)
        sign_By = np.sign(By)
        sign_By[sign_By == 0] = 1.
        ca = sign_By * (np.pi/2 - np.arcsin(Bz / Bt))
        ca = np.mod(ca, 2 * np.pi)

        if not to_radian:
            ca = ca / np.pi * 180

        var_name = 'CA_' + cs
        var = self.add_variable(var_name, configured_variables=var_config.configured_variables, configured_variable_name='CA')
        var.value = ca.reshape(ca.size, 1)
        var.label = r'$\theta$'
        var.unit = 'degree'
        var.unit_label = r'($^\circ$)'
        self[var_name] = var
        return self[var_name]

    @_validate_IMF_cs
    def add_IMF_AZ(self, cs='GSM', to_radian=False,  **kwargs):
        """
        IMF azimuthal angle in the IMF Bx-By (x-y) plane in the GSE/GSM coordinate system.
        :return:
        """
        By = kwargs['By']
        Bx = kwargs['Bx']
        Bt = np.sqrt(By**2+Bx**2)
        sign_Bx = np.sign(Bx)
        sign_Bx[sign_Bx == 0] = 1.
        az = sign_Bx * (np.pi/2 - np.arcsin(By / Bt))
        az = np.mod(az, 2 * np.pi)

        if not to_radian:
            az = az / np.pi * 180

        var_name = 'AZ_' + cs
        var = self.add_variable(var_name, configured_variables=var_config.configured_variables, configured_variable_name='CA')
        var.value = az.reshape(az.size, 1)
        var.label = r'$\phi$'
        var.unit = 'degree'
        var.unit_label = r'($^\circ$)'
        self[var_name] = var
        return self[var_name]

    @_validate_IMF_cs
    def add_IMF_EL(self, cs='GSM', to_radian=False,  **kwargs):
        """
        IMF elevation angle in the IMF Bz-Bx (x-y) plane in the GSE/GSM coordinate system.
        :return:
        """
        Bx = kwargs['Bx']
        Bz = kwargs['Bz']
        Bt = np.sqrt(Bx ** 2 + Bz ** 2)
        sign_Bz = np.sign(Bz)
        sign_Bz[sign_Bz == 0] = 1.
        el = sign_Bz * (np.pi / 2 - np.arcsin(Bx / Bt))
        el = np.mod(el, 2 * np.pi)

        if not to_radian:
            el = el / np.pi * 180

        var_name = 'EL_' + cs
        var = self.add_variable(var_name, configured_variables=var_config.configured_variables, configured_variable_name='CA')
        var.value = el.reshape(el.size, 1)
        var.label = r'$\alpha$'
        var.unit = 'degree'
        var.unit_label = r'($^\circ$)'
        self[var_name] = var
        return self[var_name]

    def add_NCF(self):
        """
        Solar wind-magnetosphere coupling function by Newell et al., JGR, 2007.
        :return:
        """
        if 'CA_GSM' not in self.keys():
            self.add_IMF_CA()

        ca = self['CA_GSM'].value
        ca = ca * np.pi / 180
        v = self['v_sw'].value
        BT = self['B_T_GSM'].value
        ncf = np.abs(v)**(4/3) * np.abs(BT)**(2/3) * np.abs(np.sin(ca*0.5))**(8/3)

        self['NCF'] = var_config.configured_variables['NCF'].clone()
        self['NCF'].value = ncf

        return self['NCF']

    @property
    def database(self):
        return self._database

    @database.setter
    def database(self, value):
        if isinstance(value, str):
            self._database = DatabaseModel(value)
        elif issubclass(value.__class__, DatabaseModel):
            self._database = value
        else:
            raise TypeError

    @property
    def facility(self):
        return self._facility

    @facility.setter
    def facility(self, value):
        if isinstance(value, str):
            self._facility = FacilityModel(value)
        elif issubclass(value.__class__, FacilityModel):
            self._facility = value
        else:
            raise TypeError
















# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLAB"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"


import datetime
import numpy as np
import requests
import bs4
import pathlib
import re
import netCDF4
import cftime

import geospacelab.toolbox.utilities.pydatetime as dttool
import geospacelab.toolbox.utilities.pylogging as mylog
from geospacelab.config import prf


class Downloader(object):

    def __init__(self, dt_fr,  dt_to, data_file_root_dir=None, force_download=False):

        self.dt_fr = dt_fr
        self.dt_to = dt_to
        self.force_download = force_download

        self.done = False
        if data_file_root_dir is None:
            self.data_file_root_dir = prf.datahub_data_root_dir / 'FMI' / 'IMAGE' / 'IE'
        else:
            self.data_file_root_dir = pathlib.Path(data_file_root_dir)

        self.url_base = "https://space.fmi.fi/image/www/data_download.php"

        self.download()

    def download(self):
        diff_days = dttool.get_diff_days(self.dt_fr, self.dt_to)
        dt0 = dttool.get_start_of_the_day(self.dt_fr)
        
        for i in range(diff_days + 1):
            starttime = dt0 + datetime.timedelta(days=i)
             
            params = {
                'starttime': starttime.strftime('%Y%m%d'),
                'length': '1440',
                'format': 'ie-index'
            }
            
            file_name = 'FMI_IMAGE_IE_' + starttime.strftime('%Y%m%d') + '.dat'
            file_dir = self.data_file_root_dir / '{:4d}'.format(starttime.year) 
            file_dir.mkdir(parents=True, exist_ok=True)
            file_path = file_dir / file_name
            
            if file_path.exists() & (not self.force_download):
                mylog.simpleinfo.info(f"The file has existed: {file_path}.")
                self.done = True
                continue
            mylog.simpleinfo.info(f'Downloading the data on {starttime.strftime("%Y-%m-%d")} ...') 
            r = requests.get(self.url_base, params=params)
            if "No data" in r.text:
                mylog.StreamLogger.warning(f"No data found on {starttime.strftime('%Y-%m-%d')}")
                self.done = (self.done or False)
                continue
            with open(file_path.with_suffix('.dat'), 'w') as f:
                f.write(r.text) 
            self.done = True
            mylog.simpleinfo.info('Done.')
        
        return


if __name__ == '__main__':
    dt1 = datetime.datetime(2009, 1, 1)
    dt2 = datetime.datetime(2009, 12, 31)
    
    Downloader(dt1, dt2)




# form_dst = {'SCent': 20, 'STens': 1, 'SYear': 1, 'SMonth': '01', 'ECent': 20, 'ETens': 1, 'EYear': 1, 'EMonth': 12, "Image Type": "GIF", "COLOR": "COLOR", "AE Sensitivity": "100", "Dst Sensitivity": "20", "Output": 'DST', "Out format": "IAGA2002", "Email": "lei.cai@oulu.fi"}


# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"


import numpy
from scipy.interpolate import interp1d
import geospacelab.toolbox.utilities.pydatetime as dttool
import geospacelab.toolbox.utilities.pybasic as basic
import datetime


def numpy_array_join_vertical(arr1, arr2):

    ndim2 = arr2.ndim
    if ndim2 == 1:
        ashape = (1, arr2.shape[0])
        arr2 = arr2.reshape(ashape)

    if arr1 is None:
        arr1 = numpy.empty((0, arr2.shape[1]))
    else:
        ndim1 = arr1.ndim
        if ndim1 == 1:
            ashape = (1, arr1.shape[0])
            arr1 = arr1.reshape(ashape)
    newarr = numpy.vstack((arr1, arr2))
    return newarr


def numpy_array_join_horizontal(arr1, arr2):
    ndim2 = arr2.ndim
    if ndim2 == 1:
        ashape = (arr2.shape[0], 1)
        arr2 = arr2.reshape(ashape)

    if arr1 is None:
        arr1 = numpy.empty((arr2.shape[0], 0))
    else:
        ndim1 = arr1.ndim
        if ndim1 == 1:
            ashape = (arr1.shape[0], 1)
            arr1 = arr1.reshape(ashape)
    newarr = numpy.hstack((arr1, arr2))
    return newarr


def numpy_array_self_mask(data, conditions=None):
    # conditions should be a list
    if conditions is None:
        return data
    for condition in conditions:
        if isinstance(condition, (float, int, str)):
            data = numpy.ma.masked_where(data == condition, data)
        elif isinstance(condition, dict):
            for key, value in condition.items():
                if key == 'between':
                    data = numpy.ma.masked_where((data >= value[0]) & (data <= value[1]), data)
                elif key == 'out':
                    data = numpy.ma.masked_where((data < value[0]) | (data > value[1]), data)
                elif key == 'greater':
                    data = numpy.ma.masked_where(data > value, data)
                elif key == 'smaller':
                    data = numpy.ma.masked_where(data < value, data)
                elif key == 'eq':
                    if not isinstance(value, list):
                        value = [value]
                    for v in value:
                        data = numpy.ma.masked_where(data == value, data)
                elif key == 'neq':
                    data = numpy.ma.masked_where(data != value, data)
    return data


def data_resample(
        x=None, y=None, xtype=None, xres=None, xresscale=1.1,
        method='Null',  # Null - insert NaN, 'linear', 'cubic', ... (interpolation method)
        axis=0, forward=True, depth=0.
):

    x1 = x
    if xtype == 'datetime':
        # dt0 = datetime.datetime(1970, 1, 1)
        sectime, dt0 = dttool.convert_datetime_to_sectime(x1)
        x1 = sectime

    diff_x1 = numpy.diff(x1.flatten())
    if xres is None:
        xres = numpy.median(diff_x1)
    inds = numpy.where(diff_x1 > xres * xresscale)[0]

    if len(inds) == 0:
        return x, y
    if depth > 10:
        return x, y

    inds = [i+1 for i in inds]
    # for x
    value = []
    for ind in inds:
        res = xres
        if xtype == 'datetime':
            res = datetime.timedelta(seconds=xres)
        if forward:
            value.append(x[ind - 1] + res/10)
        else:
            value.append(x[ind] - res/10)
    xnew = numpy.insert(x, inds, value, axis=axis)

    # for y
    if method == 'Null':
        value = numpy.nan
        ynew = numpy.insert(y, inds, value, axis=axis)
    else:
        ifunc = interp1d(x1.flatten(), y, kind=method, axis=axis)
        x_p = xnew
        if xtype=='datetime':
            x_p, dt0 = dttool.convert_datetime_to_sectime(xnew)
        ynew = ifunc(x_p.flatten())

    if method == 'Null':
        xnew, ynew = data_resample(
        x=xnew, y=ynew, xtype=xtype, xres=xres, xresscale=xresscale,
        method='Null',  # Null - insert NaN, 'linear', 'cubic', ... (interpolation method)
        axis=axis, forward=False, depth=depth+1
)

    return xnew, ynew


def regridding_2d_xgaps(
    x, y, z,
    xtype=None, xres=None):
    
    x1 = x.flatten()
    if xtype == 'datetime':
        # dt0 = datetime.datetime(1970, 1, 1)
        sectime, dt0 = dttool.convert_datetime_to_sectime(x1)
        x1 = sectime
    
    diff_x = numpy.diff(x1)
    if xres is None:
        xres = numpy.median(diff_x)
        
    if xtype == 'datetime' and basic.isnumeric(xres):
        xres = datetime.timedelta(seconds=xres)
        
    xx = numpy.hstack(((x.flatten() - xres / 2)[:, numpy.newaxis], (x.flatten() + xres / 2)[:, numpy.newaxis]))
    xnew = xx.flatten()
    
    if len(y.shape) == 1 or 1 in y.shape:
        ynew = y
    else:
        yy = numpy.hstack((y, y))
        ynew = yy.reshape((xnew.shape[0], y.shape[1]))
    
    zz = numpy.hstack((z, z))
    znew = zz.reshape((xnew.shape[0], z.shape[1]))
    znew[1::2, :] = numpy.nan
    znew = znew[:-1, :]
    
    return xnew, ynew, znew
         

def data_resample_2d(
        x=None, y=None, z=None, xtype=None, xres=None, xresscale=3,
        method='Null',  # Null - insert NaN, 'linear', 'cubic', ... (interpolation method)
        axis=0, forward=True, depth=0
):

    x1 = x
    if xtype == 'datetime':
        # dt0 = datetime.datetime(1970, 1, 1)
        sectime, dt0 = dttool.convert_datetime_to_sectime(x1)
        x1 = sectime

    diff_x1 = numpy.diff(x1.flatten())
    if xres is None:
        xres = numpy.median(diff_x1)
    inds = numpy.where(diff_x1 > xres * xresscale)[0]

    if len(inds) == 0:
        return x, y, z

    if depth > 10:
        return x, y, z

    inds = [i+1 for i in inds]
    # for x
    value = []
    for ind in inds:
        res = xres
        if xtype == 'datetime':
            res = datetime.timedelta(seconds=xres)
        if forward:
            value.append(x[ind - 1] + res/10)
        else:
            value.append(x[ind] - res/10)
    xnew = numpy.insert(x, inds, value, axis=axis)

    # for y
    method_y = 'nearest'
    ifunc = interp1d(x1.flatten(), y, kind=method_y, axis=axis, fill_value="extrapolate")
    x_p = xnew
    if xtype == 'datetime':
        x_p, dt0 = dttool.convert_datetime_to_sectime(xnew)
    ynew = ifunc(x_p.flatten())

    # for z
    if method == 'Null':
        value = numpy.nan
        znew = numpy.insert(z, inds, value, axis=axis)
    else:
        ifunc = interp1d(x1.flatten(), z, kind=method, axis=axis)
        x_p = xnew
        if xtype == 'datetime':
            x_p, dt0 = dttool.convert_datetime_to_sectime(xnew)
        znew = ifunc(x_p.flatten())
    if method == 'Null':
        xnew, ynew, znew = data_resample_2d(
            x=xnew, y=ynew, z=znew, xtype=xtype, xres=xres, xresscale=xresscale,
            method='Null',  # Null - insert NaN, 'linear', 'cubic', ... (interpolation method)
            axis=axis, forward=False, depth=depth+1
        )

    return xnew, ynew, znew



# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"

import datetime
import numpy as np

from geospacelab.datahub.sources.madrigal.isr.millstonehill.basic.loader import Loader as default_Loader
from geospacelab.datahub.sources.madrigal.isr.millstonehill.downloader import Downloader as default_Downloader
import geospacelab.datahub.sources.madrigal.isr.millstonehill.basic.variable_config as var_config
from geospacelab.datahub.sources.madrigal import madrigal_database
from geospacelab.config import prf
from geospacelab import datahub
from geospacelab.datahub import SiteModel, DatabaseModel, FacilityModel
import geospacelab.toolbox.utilities.pybasic as basic
import geospacelab.toolbox.utilities.pydatetime as dttool
import geospacelab.toolbox.utilities.pylogging as mylog

default_dataset_attrs = {
    'kind': 'sourced',
    'database': madrigal_database,
    'facility': 'MillstoneHillISR',
    'exp_name_pattern': '',
    'exp_check': False,
    'data_file_type': '',
    'data_file_ext': 'hdf5',
    'data_root_dir': prf.datahub_data_root_dir / 'Madrigal' / 'MillstoneHill_ISR',
    'antenna': '',
    'pulse_code': '',
    'pulse_length': 0,
    'allow_download': True,
    'status_control': False,
    'residual_control': False,
    'beam_location': True,
    'data_search_recursive': True,
    'label_fields': ['database', 'facility', 'site', 'antenna', 'pulse_code', 'pulse_length'],
}

default_variable_names = [
    'DATETIME', 'AZ', 'AZ1', 'AZ2', 'EL', 'EL1', 'EL2', 'PULSE_LENGTH',
    'T_SYS', 'POWER_NORM', 'P_Tx', 'MODE_TYPE', 'POWER_LENGTH_F',
    'LAG_SPACING', 'IPP', 'f_Tx', 'v_PHASE_Tx', 'v_PHASE_Tx_err',
    'SCAN_TYPE', 'CYCN', 'POSN', 'RANGE_RES', 'RANGE',
    'SNR', 'RESIDUAL', 'STATUS', 'FIT_TYPE', 'FPI_QUALITY',
    'ACF_NORM', 'ACF_NORM_ERR', 'n_pp', 'n_pp_err', 'n_e',
    'n_e_err', 'T_i', 'T_i_err', 'T_r', 'T_r_err', 'T_e', 'T_e_err',
    'nu_i', 'nu_i_err', 'v_i_los', 'v_i_los_err', 'comp_H_p',
    'comp_H_p_err', 'comp_mix', 'comp_mix_err', 'v_DOP_los', 'v_DOP_los_err',
    'HEIGHT'
]

# default_data_search_recursive = True

default_attrs_required = ['antenna', 'pulse_code', 'pulse_length']


class Dataset(datahub.DatasetSourced):
    def __init__(self, **kwargs):
        kwargs = basic.dict_set_default(kwargs, **default_dataset_attrs)

        super().__init__(**kwargs)

        self.database = kwargs.pop('database', '')
        self.facility = kwargs.pop('facility', '')
        self.site = kwargs.pop('site', MillstoneHillSite('MillstoneHill'))
        self.antenna = kwargs.pop('antenna', '')
        self.experiment = kwargs.pop('experiment', '')
        self.exp_name_pattern = kwargs.pop('exp_name_pattern', '')
        self.exp_check = kwargs.pop('exp_check', False)
        self.pulse_code = kwargs.pop('pulse_code', '')
        self.pulse_length = kwargs.pop('pulse_length', '')
        self.data_file_type = kwargs.pop('data_file_type', '')
        self.affiliation = kwargs.pop('affiliation', '')
        self.allow_download = kwargs.pop('allow_download', True)
        self.beam_location = kwargs.pop('beam_location', True)
        self.metadata = None

        allow_load = kwargs.pop('allow_load', True)
        self.status_control = kwargs.pop('status_control', False)
        self.residual_control = kwargs.pop('residual_control', False)

        # self.config(**kwargs)

        if self.loader is None:
            self.loader = default_Loader
        if self.downloader is None:
            self.downloader = default_Downloader

        self._set_default_variables(
            default_variable_names,
            configured_variables=var_config.configured_variables
        )

        self._validate_attrs()

        if allow_load:
            self.load_data()

    def _validate_attrs(self):
        for attr_name in default_attrs_required:
            attr = getattr(self, attr_name)
            if not str(attr):
                mylog.StreamLogger.warning("The parameter {} is required before loading data!".format(attr_name))

        if self.exp_check:
            self.download_data()
            self.exp_check = False

    def label(self, **kwargs):
        label = super().label()
        return label

    def load_data(self, **kwargs):
        self.check_data_files(**kwargs)

        for file_path in self.data_file_paths:
            load_obj = self.loader(file_path, antenna=self.antenna, pulse_code=self.pulse_code, pulse_length=self.pulse_length)

            for var_name in self._variables.keys():
                self._variables[var_name].join(load_obj.variables[var_name])

                if var_name in ['n_pp', 'n_pp_err', 'n_e',
                                'n_e_err', 'T_i', 'T_i_err', 'T_r', 'T_r_err', 'T_e', 'T_e_err',
                                'nu_i', 'nu_i_err', 'v_i_los', 'v_i_los_err', 'comp_H_p',
                                'comp_H_p_err', 'comp_mix', 'comp_mix_err', 'v_DOP_los', 'v_DOP_los_err', ]:
                    if self.pulse_code == 'single pulse':
                        self._variables[var_name].visual.axis[1].lim = [90, 600]

            self.antenna = load_obj.metadata['antenna']
            self.pulse_code: str = load_obj.metadata['pulse_code']
            self.pulse_length = load_obj.metadata['pulse_length']
            self.metadata = load_obj.metadata
        if self.beam_location:
            self.calc_lat_lon()
            # self.select_beams(field_aligned=True)
        if self.time_clip:
            self.time_filter_by_range()
        if self.status_control:
            self.status_mask()
        if self.residual_control:
            self.residual_mask()

    def outlier_mask(self, condition, fill_value=None):
        """
        Mask outliers of the 2D variables (Ne, Te, Ti, ...) depending on condition.
        :param condition: a bool array with the same dimension as the 2D variables
        :type condition: np.ndarray, bool
        :param fill_value: the value filled as the outlier, [np.nan]
        :type fill_value: np.nan or float
        """
        var_2d_names = [
            'n_pp', 'n_pp_err', 'n_e',
            'n_e_err', 'T_i', 'T_i_err', 'T_r', 'T_r_err',
            'nu_i', 'nu_i_err', 'v_i_los', 'v_i_los_err', 'comp_H_p',
            'comp_H_p_err', 'comp_mix', 'comp_mix_err', 'v_DOP_los', 'v_DOP_los_err',
        ]
        for key in self.keys():
            if key in var_2d_names:
                self[key].value[condition] = np.nan

    def status_mask(self, bad_status=None):
        """
        Mask the 2D variables depending on status.
        :param bad_status: a list of the status flags, [2, 3].
        :type bad_status: list
        """
        if bad_status is None:
            bad_status = [2, 3]
        for flag in bad_status:
            condition = self['STATUS'].value == flag
            self.outlier_mask(condition)

    def residual_mask(self, residual_lim=None):
        """
        Mask the 2D variables depending on residual
        :param residual_lim: the lower limit of the bad residual, [10].
        :type residual_lim: float
        """
        if residual_lim is None:
            residual_lim = 10

        condition = self['RESIDUAL'].value > residual_lim
        self.outlier_mask(condition)

    def select_beams(self, field_aligned=False, az_el_pairs=None, error_az=2, error_el=2):
        if field_aligned:
            if az_el_pairs is not None:
                raise AttributeError("The parameters field_aligned and az_el_pairs cannot be set at the same time!")
            if self.site != 'UHF':
                raise AttributeError("Only UHF can be applied.")

        az = self['AZ'].value.flatten() % 360.
        el = self['EL'].value.flatten()
        if field_aligned:
            inds = np.where(((np.abs(az - 188.6) <= error_az) & (np.abs(el-77.7) <= error_el)))[0]
            if not list(inds):
                mylog.StreamLogger.error("No field-aligned beams found!")
                raise ValueError

        elif isinstance(az_el_pairs, list):
            inds = []
            for az1, el1 in az_el_pairs:
                az1 = az1 % 360.
                ind_1 = np.where(((np.abs(az - az1) <= error_az) & (np.abs(el-el1) <= error_el)))[0]
                ind_2 = np.where(((np.abs(az - 360. - az1) <= error_az) & (np.abs(el-el1) <= error_el)))[0] 
                ind_1 = np.append(ind_1, ind_2)
                if not list(ind_1):
                    mylog.StreamLogger.error("Cannot find the beam with az={:f} and el={:f}".format(az1, el1))
                    raise ValueError
                inds.extend(ind_1)
            inds = np.sort(np.unique(np.array(inds))).tolist()
        else:
            raise ValueError
        self.time_filter_by_inds(inds)

    def calc_lat_lon(self, AACGM=True):
        from geospacelab.cs import LENUSpherical
        az = self['AZ'].value
        el = self['EL'].value
        range = self['RANGE'].value
        az = np.tile(az, (1, range.shape[1]))  # make az, el, range in the same shape
        el = np.tile(el, (1, range.shape[1]))

        # az = np.array([0, 90, 180, 270])
        # el = np.array([0, 45, 90, 45])
        # range = np.array([0, 100, 200, 100])
        cs_LENU = LENUSpherical(coords={'az': az, 'el': el, 'range': range},
                                lat_0=self.site.location['GEO_LAT'],
                                lon_0=self.site.location['GEO_LON'],
                                height_0=self.site.location['GEO_ALT'])
        cs_new = cs_LENU.to_GEO()
        configured_variables = var_config.configured_variables

        var = self.add_variable(var_name='GEO_LAT', value=cs_new['lat'],
                                configured_variables=configured_variables)
        var = self.add_variable(var_name='GEO_LON', value=cs_new['lon'],
                                configured_variables=configured_variables)
        var = self.add_variable(var_name='GEO_ALT', value=cs_new['height'],
                                configured_variables=configured_variables)

        if AACGM:
            cs_new.ut = self['DATETIME'].value
            cs_new = cs_new.to_AACGM()
        var = self.add_variable(var_name='AACGM_LAT', value=cs_new['lat'],
                                configured_variables=configured_variables)
        var = self.add_variable(var_name='AACGM_LON', value=cs_new['lon'],
                                configured_variables=configured_variables)
        # var = self.add_variable(var_name='AACGM_ALT', value=cs_new['height'])
        pass

    def search_data_files(self, recursive=True, **kwargs):
        dt_fr = self.dt_fr
        dt_to = self.dt_to
        diff_days = dttool.get_diff_days(dt_fr, dt_to)
        day0 = dttool.get_start_of_the_day(dt_fr)
        for i in range(diff_days + 1):
            thisday = day0 + datetime.timedelta(days=i)
            initial_file_dir = self.data_root_dir / thisday.strftime('%Y') / thisday.strftime('%Y%m%d')

            if self.data_file_type == 'combined':
                file_patterns = ['MillstoneHill', self.data_file_type, thisday.strftime('%Y%m%d')]
            else:
                file_patterns = ['MillstoneHill', self.antenna, self.pulse_code.replace(' ', '_'), thisday.strftime('%Y%m%d')]

            # remove empty str
            file_patterns = [pattern for pattern in file_patterns if str(pattern)]

            search_pattern = '*' + '*'.join(file_patterns) + '*'
            if str(self.exp_name_pattern):
                search_pattern = '*' + self.exp_name_pattern.replace(' ', '-') + '*/' + search_pattern
                recursive = False

            done = super().search_data_files(
                initial_file_dir=initial_file_dir, search_pattern=search_pattern, recursive=recursive)


            # Validate file paths

            if not done and self.allow_download:
                done = self.download_data()
                if done:
                    done = super().search_data_files(
                        initial_file_dir=initial_file_dir, search_pattern=search_pattern, recursive=recursive)
                else:
                    print('The requested experiment does not exist in the online database!')
                    raise ValueError

            if len(done) > 1:
                if str(self.exp_name_pattern):
                    mylog.StreamLogger.error(
                        "Multiple data files detected! Check the files:")
                else:
                    mylog.StreamLogger.error(
                        "Multiple data files detected!" +
                        "Specify the experiment name by the keyword 'exp_name_pattern' if possible.")
                for fp in done:
                    mylog.simpleinfo.info(fp)
                raise KeyError

        return done

    def download_data(self):
        download_obj = self.downloader(
            dt_fr=self.dt_fr, dt_to=self.dt_to,
            data_file_root_dir=self.data_root_dir,
            file_type=self.data_file_type,
            exp_name_pattern=self.exp_name_pattern, dry_run=self.exp_check)
        return download_obj.done

    @property
    def database(self):
        return self._database

    @database.setter
    def database(self, value):
        if isinstance(value, str):
            self._database = DatabaseModel(value)
        elif issubclass(value.__class__, DatabaseModel):
            self._database = value
        else:
            raise TypeError

    @property
    def facility(self):
        return self._facility

    @facility.setter
    def facility(self, value):
        if isinstance(value, str):
            self._facility = FacilityModel(value)
        elif issubclass(value.__class__, FacilityModel):
            self._facility = value
        else:
            raise TypeError

    @property
    def site(self):
        return self._site

    @site.setter
    def site(self, value):
        if isinstance(value, str):
            self._site = MillstoneHillSite(value)
        elif issubclass(value.__class__, SiteModel):
            self._site = value
        else:
            raise TypeError


class MillstoneHillSite(SiteModel):
    def __new__(cls, str_in, **kwargs):
        obj = super().__new__(cls, str_in, **kwargs)
        return obj

    def __init__(self, str_in, **kwargs):
        self.name = 'MillstoneHill'
        self.location = {
            'GEO_LAT': 42.619,
            'GEO_LON': 288.51,
            'GEO_ALT': 0.146
        }















# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"


import datetime
import requests
import bs4
import re

import geospacelab.toolbox.utilities.pydatetime as dttool
import geospacelab.toolbox.utilities.pylogging as mylog
from geospacelab.config import prf


class Downloader(object):
    def __init__(self, dt_fr,  dt_to, res='1min', new_omni=True, data_file_root_dir=None, version=None):
        if res == '1h':
            new_omni = True
        self.dt_fr = dt_fr
        self.dt_to = dt_to
        self.res = res
        self.new_omni = new_omni
        self.version = version
        self.done = False
        if data_file_root_dir is None:
            self.data_file_root_dir = prf.datahub_data_root_dir / "CDAWeb" / 'OMNI'
        else:
            self.data_file_root_dir = data_file_root_dir

        self.url_base = "https://cdaweb.gsfc.nasa.gov/pub/data/omni/omni_cdaweb/"

        self.download()

    def download(self):
        if self.res in ['1min', '5min']:
            self.download_high_res_omni()
        elif self.res == '1h':
            self.download_low_res_omni()

    def download_high_res_omni(self):
        if self.new_omni:
            omni_type = "hro2"
            omni_dir_name = 'OMNI2'
        else:
            omni_type = "hro"
            omni_dir_name = "OMNI"

        num_month = (self.dt_to.year - self.dt_fr.year) * 12 + self.dt_to.month - self.dt_fr.month + 1
        dt0 = datetime.datetime(self.dt_fr.year, self.dt_fr.month, 1)
        for nm in range(num_month):
            dt1 = dttool.get_next_n_months(dt0, nm)
            url = self.url_base + omni_type + '_' + self.res + '/' + '{:4d}'.format(dt1.year) + '/'

            r = requests.get(url)

            soup = bs4.BeautifulSoup(r.text, 'html.parser')
            a_tags = soup.find_all('a', href=True)
            hrefs = []
            for a_tag in a_tags:
                href = a_tag['href']
                pattern = omni_type + '_' + self.res + '_' + dt1.strftime("%Y%m%d")
                if pattern in href:
                    hrefs.append(href)
            if len(hrefs) == 0:
                mylog.StreamLogger.info("Cannot find the queried data file!")
                return
            if len(hrefs) > 1:
                mylog.StreamLogger.warning("Find multiple matched files!")
                print(hrefs)
                return

            href = hrefs[0]
            ma = re.search('v[0-5][0-9]', href)
            version = ma.group(0)
            if self.version is None:
                self.version = version
            elif self.version != version:
                mylog.StreamLogger.info("Cannot find the queried data file! Version={}.".format(version))

            r_file = requests.get(url + href, allow_redirects=True)
            file_name = href
            file_path = self.data_file_root_dir / (omni_dir_name + '_high_res_' + self.res)
            file_path = file_path / '{:4d}'.format(dt1.year) / file_name
            if file_path.is_file():
                mylog.simpleinfo.info(
                    "The file {} exists in the directory {}.".format(file_path.name, file_path.parent.resolve()))
            else:
                file_path.parent.resolve().mkdir(parents=True, exist_ok=True)
                with open(file_path, "wb") as omni:
                    mylog.simpleinfo.info(
                        "Downloading {} to the directory {} ...".format(file_path.name, file_path.parent.resolve())
                    )
                    omni.write(r_file.content)
                    mylog.simpleinfo.info("Done")
            self.done = True

    def download_low_res_omni(self):
        raise NotImplemented


def test():
    dt_fr = datetime.datetime(2020, 3, 4)
    dt_to = datetime.datetime(2020, 5, 3)
    download_obj = Downloader(dt_fr, dt_to)
    pass


if __name__ == "__main__":
    test()





class Panel(object):
    def __init__(self):
        pass

    def add_line(self):
        pass

    def add_image(self):
        pass

    def add_pcolor(self):
        pass

    def add_scatter(self):
        pass

    



#
#
# def sun_position(doy, time, glat, glon, height, **kwargs):
#
#
#
# subroutine soco (ld,t,flat,Elon,height,
#      &          DECLIN, ZENITH, SUNRSE, SUNSET)
# c--------------------------------------------------------------------
# c       s/r to calculate the solar declination, zenith angle, and
# c       sunrise & sunset times  - based on Newbern Smith's algorithm
# c       [leo mcnamara, 1-sep-86, last modified 16-jun-87]
# c       {dieter bilitza, 30-oct-89, modified for IRI application}
# c
# c in:   ld      local day of year
# c       t       local hour (decimal)
# c       flat    northern latitude in degrees
# c       elon    east longitude in degrees
# c		height	height in km
# c
# c out:  declin      declination of the sun in degrees
# c       zenith      zenith angle of the sun in degrees
# c       sunrse      local time of sunrise in hours
# c       sunset      local time of sunset in hours
# c-------------------------------------------------------------------
# c
#         common/const/dtr,pi     /const1/humr,dumr
# c amplitudes of Fourier coefficients  --  1955 epoch.................
#         data    p1,p2,p3,p4,p6 /
#      &  0.017203534,0.034407068,0.051610602,0.068814136,0.103221204 /
# c
# c s/r is formulated in terms of WEST longitude.......................
#         wlon = 360. - Elon
# c
# c time of equinox for 1980...........................................
#         td = ld + (t + Wlon/15.) / 24.
#         te = td + 0.9369
# c
# c declination of the sun..............................................
#         dcl = 23.256 * sin(p1*(te-82.242)) + 0.381 * sin(p2*(te-44.855))
#      &      + 0.167 * sin(p3*(te-23.355)) - 0.013 * sin(p4*(te+11.97))
#      &      + 0.011 * sin(p6*(te-10.41)) + 0.339137
#         DECLIN = dcl
#         dc = dcl * dtr
# c
# c the equation of time................................................
#         tf = te - 0.5
#         eqt = -7.38*sin(p1*(tf-4.)) - 9.87*sin(p2*(tf+9.))
#      &      + 0.27*sin(p3*(tf-53.)) - 0.2*cos(p4*(tf-17.))
#         et = eqt * dtr / 4.
# c
#         fa = flat * dtr
#         phi = humr * ( t - 12.) + et
# c
#         a = sin(fa) * sin(dc)
#         b = cos(fa) * cos(dc)
#         	cosx = a + b * cos(phi)
#         	if(abs(cosx).gt.1.) cosx=sign(1.,cosx)
#         zenith = acos(cosx) / dtr
# c
# c calculate sunrise and sunset times --  at the ground...........
# c see Explanatory Supplement to the Ephemeris (1961) pg 401......
# c sunrise at height h metres is at...............................
# 		h=height*1000.
#         chih = 90.83 + 0.0347 * sqrt(h)
# c this includes corrections for horizontal refraction and........
# c semi-diameter of the solar disk................................
#         ch = cos(chih * dtr)
#         cosphi = (ch -a ) / b
# c if abs(secphi) > 1., sun does not rise/set.....................
# c allow for sun never setting - high latitude summer.............
#         secphi = 999999.
#         if(cosphi.ne.0.) secphi = 1./cosphi
#         sunset = 99.
#         sunrse = 99.
#         if(secphi.gt.-1.0.and.secphi.le.0.) return
# c allow for sun never rising - high latitude winter..............
#         sunset = -99.
#         sunrse = -99.
#         if(secphi.gt.0.0.and.secphi.lt.1.) return
# c
#         	cosx = cosphi
#         	if(abs(cosx).gt.1.) cosx=sign(1.,cosx)
#         phi = acos(cosx)
#         et = et / humr
#         phi = phi / humr
#         sunrse = 12. - phi - et
#         sunset = 12. + phi - et
#         if(sunrse.lt.0.) sunrse = sunrse + 24.
#         if(sunset.ge.24.) sunset = sunset - 24.
# c special case sunrse > sunset
#         if(sunrse.gt.sunset) then
#         	sunx=sign(99.0,flat)
#         	if(ld.gt.91.and.ld.lt.273) then
#         		sunset = sunx
#         		sunrse = sunx
#         	else
#         		sunset = -sunx
#         		sunrse = -sunx
#         	endif
#         	endif
# c
#         return
#         end
# c


# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"

import re
import numpy as np
import datetime
import geospacelab.toolbox.utilities.pydatetime as dttool


class Loader(object):

    def __init__(self, file_path, file_type='txt', version='v01', direct_load=True):
        self.file_path = file_path
        self.file_type = file_type
        self.version = version
        self.variables = {}
        self.done = False
        if direct_load:
            self.load()

    def load(self):

        if self.version == 'v01':
            raise ValueError
        elif 'v02' in self.version:
            self.load_v02()
        else:
            raise NotImplementedError


    def load_v02(self):
        with open(self.file_path, 'r') as f:
            text = f.read()
            results = re.findall(
            r"^(\d{4}-\d{2}-\d{2}\s*\d{2}\:\d{2}\:\d{2}\.\d{3})\s*(\w{3})\s*" 
            + r"([\-\d.]+)\s*([\-\d.]+)\s*([\-\d.]+)\s*([\-\d.]+)\s*([\-\d.]+)\s*([+\-\d.Ee]+)\s*([+\-\d.Ee]+)\s*([\-\d.]+)\s*([\-\d.]+)",
            text,
            re.M)
            results = list(zip(*results))
            dts = [datetime.datetime.strptime(dtstr + '000', "%Y-%m-%d %H:%M:%S.%f") for dtstr in results[0]]
            if results[1][0] == 'GPS':
                t_gps = [(dt - dttool._GPS_DATETIME_0).total_seconds() for dt in dts]
                dts = dttool.convert_gps_time_to_datetime(t_gps, weeks=None)
            num_rec = len(dts)
            self.variables['SC_DATETIME'] = np.array(dts).reshape(num_rec, 1)
            self.variables['SC_GEO_ALT'] = np.array(results[2]).astype(np.float32).reshape(num_rec, 1) * 1e-3   # in km
            self.variables['SC_GEO_LON'] = np.array(results[3]).astype(np.float32).reshape(num_rec, 1)
            self.variables['SC_GEO_LAT'] = np.array(results[4]).astype(np.float32).reshape(num_rec, 1)
            self.variables['SC_GEO_LST'] = np.array(results[5]).astype(np.float32).reshape(num_rec, 1)
            self.variables['SC_ARG_LAT'] = np.array(results[6]).astype(np.float32).reshape(num_rec, 1)
            self.variables['rho_n'] = np.array(results[7]).astype(np.float32).reshape(num_rec, 1)
            self.variables['rho_n_MEAN'] = np.array(results[8]).astype(np.float32).reshape(num_rec, 1) 
            self.variables['FLAG'] = np.array(results[9]).astype(np.float32).reshape(num_rec, 1) 
            self.variables['FLAG_MEAN'] = np.array(results[10]).reshape(num_rec, 1)


# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

from geospacelab.datahub import FacilityModel


class DMSPFacility(FacilityModel):
    def __new__(cls, str_in, **kwargs):
        obj = super().__new__(cls, str_in, **kwargs)
        return obj


dmsp_facility = DMSPFacility('DMSP')
dmsp_facility.url = ''
dmsp_facility.category = 'Madrigal online/DMSP'
dmsp_facility.Notes = ''



# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"


import datetime
import numpy as np
import cftime
import requests
import bs4
import pathlib
import re
import netCDF4 as nc
import pandas as pd
import ftplib
from contextlib import closing

import geospacelab.toolbox.utilities.pydatetime as dttool
import geospacelab.toolbox.utilities.pylogging as mylog
from geospacelab.config import prf
import geospacelab.datahub.sources.supermag.supermag_api as smapi

var_name_dict = {
    'mlat': 'AACGM_LAT',
    'mlon': 'AACGM_LON',
    'mlt': 'AACGM_MLT',
    'sza': 'SZA',
    'decl': 'DECLINATION',
    'N_GEO': 'N_GEO',
    'E_GEO': 'E_GEO',
    'Z_GEO': 'Z_GEO',
    'N_MAG': 'N_MAG',
    'E_MAG': 'E_MAG',
    'Z_MAG': 'Z_MAG',
}


class Downloader(object):

    def __init__(
            self,
            dt_fr: datetime.datetime,
            dt_to: datetime.datetime,
            user_id: str = None,
            site_name: str = None,
            force_download=False,
            baseline='all',
            **kwargs
    ):

        self.data_file_root_dir = prf.datahub_data_root_dir / 'SuperMAG' / 'sites'
        self.dt_fr = dt_fr
        self.dt_to = dt_to
        if user_id is None:
            self.user_id = prf.user_config['datahub']['supermag']['username']
        self.site_name = site_name
        self.baseline = baseline

        self.force_download = force_download
        self.done = False

        self.download()


    def download(self):
        diff_month = dttool.get_diff_months(self.dt_fr, self.dt_to)
        for nm in range(diff_month+1):
            this_month = dttool.get_next_n_months(self.dt_fr, nm)
            now = datetime.datetime.now()
            if this_month.year == now.year and this_month.month == now.month:
                self.force_download = True
            next_month = dttool.get_next_n_months(self.dt_fr, nm + 1)
            extent = (next_month - this_month).total_seconds()
            flags = ['MLT', 'MAG', 'GEO', 'DECL', 'SZA',]
            if self.baseline == 'DELTA':
                flags.append('DELTA')
            elif self.baseline in ['all', 'yearly', 'none']:
                bstr = f"baseline={self.baseline}"
                flags.append(bstr)
            flags = ','.join(flags)
            file_name = '_'.join([
                'SuperMAG',
                self.site_name,
                this_month.strftime('%Y%m'),
                'baseline',
                self.baseline.lower()
            ]) + '.nc'
            file_path = self.data_file_root_dir / self.site_name / this_month.strftime('%Y') / file_name
            if file_path.is_file() and not self.force_download:
                mylog.StreamLogger.info(f'The requested data file already exists! See "{file_path}".')
                self.done = True
                continue
            mylog.simpleinfo.info(f"Downloading the {self.site_name} data from SuperMAG ...")
            (status, sm_data) = smapi.SuperMAGGetData(
                self.user_id,
                this_month,
                extent,
                flags,
                self.site_name,
                FORMAT='list')

            if status == 1:
                self.save_to_nc(sm_data, file_path)
            else:
                self.done = False
                mylog.StreamLogger.error(f'The requested data cannot be downloaded!')
                return
            self.done = True

    def save_to_nc(self, sm_data, file_path):
        def sm_t_to_datetime(tval):
            jd = (tval / 86400.0) + 2440587.5
            timestamp = pd.to_datetime(jd, unit='D', origin='julian')
            t = timestamp.to_pydatetime()       # format YYYY-MM-DD HH:MM:SS.ssssss
            if t.microsecond >= 500_000:
                t += datetime.timedelta(seconds=1)
            return t.replace(microsecond=0)

        res = list(map(lambda x: list(x.values()), sm_data))
        keys = list(sm_data[0].keys())
        res = list(zip(*res))

        keys_1d = ['tval', 'mlat', 'mlon', 'mlt', 'decl', 'sza', 'N', 'E', 'Z']

        data = {}
        for k in keys_1d:
            id = keys.index(k)
            if k not in ['N', 'E', 'Z']:
                data[k] = np.array(res[id])
                continue

            rr = list(map(lambda x: list(x.values()), res[id]))
            rr = list(zip(*rr))

            for ik, kk in enumerate(res[id][0].keys()):
                if kk == 'geo':
                    name_key = k + '_GEO'
                elif kk == 'nez':
                    name_key = k + '_MAG'
                else:
                    raise NotImplementedError
                data[name_key] = np.array(rr[ik])

        dts = np.array([sm_t_to_datetime(tval) for tval in data['tval']])

        file_path.parent.resolve().mkdir(parents=True, exist_ok=True)
        fnc = nc.Dataset(file_path, 'w')
        fnc.title = f"SuperMAG magnetometer at {self.site_name} (Daily)"
        fnc.site = self.site_name
        fnc.start_time = dttool.get_start_of_the_day(dts[0]).strftime("%Y-%m-%d %H:%M:%S")
        fnc.end_time = dttool.get_end_of_the_day(dts[-1]).strftime("%Y-%m-%d %H:%M:%S")
        fnc.GEO_LAT = sm_data[0]['glat']
        fnc.GEO_LON = sm_data[0]['glon']
        fnc.time_res = sm_data[0]['ext']
        num_rows = len(dts)
        time_array = np.array(cftime.date2num(dts, units='seconds since 1970-01-01 00:00:00.0'))
        fnc.createDimension('UNIX_TIME', num_rows)
        time = fnc.createVariable('UNIX_TIME', np.float64, ('UNIX_TIME',))
        time.units = 'seconds since 1970-01-01 00:00:00.0'
        time[::] = time_array[::]

        for k, v in var_name_dict.items():
            var = data[k]
            var_nc = fnc.createVariable(v, np.float32, ('UNIX_TIME', ))
            var_nc[::] = var[::]
        mylog.StreamLogger.info(
            "The requested SuperMAG magnetometer data has been saved in the file {}.".format(file_path))
        fnc.close()
        self.done = True


if __name__ == "__main__":
    dt_fr1 = datetime.datetime(2016, 3, 15, 15)
    dt_to1 = datetime.datetime(2016, 3, 15, 22)
    Downloader(dt_fr1, dt_to1, site_name='SKT', force_download=True)






import numpy as np
import netCDF4 as nc
import datetime
import pathlib
import cftime

from sscws.sscws import SscWs
from sscws.coordinates import CoordinateComponent, CoordinateSystem, \
    SurfaceGeographicCoordinates

import geospacelab.cs as gsl_cs
import geospacelab.toolbox.utilities.pydatetime as dttool


def sat_orbits(
        dt_fr: datetime.datetime = None, dt_to: datetime.datetime = None,
        sat_id='f18', to_AACGM='on',
):
    # initialization
    ssc = SscWs()
    sat = 'dmsp' + sat_id.lower()
    dt_fr_str = dt_fr.strftime('%Y-%m-%dT%H:%M:%SZ')
    dt_to_str = dt_to.strftime('%Y-%m-%dT%H:%M:%SZ')

    result = ssc.get_locations(
        [sat],
        [dt_fr_str, dt_to_str], [CoordinateSystem.GEO, CoordinateSystem.GSE]
    )

    if not list(result['Data']):
        return None

    data = result['Data'][0]
    coords = data['Coordinates'][0]
    coords_gse = data['Coordinates'][1]
    dts = data['Time']

    coords_in = {'x': coords['X'] / 6371.2, 'y': coords['Y'] / 6371.2, 'z': coords['Z'] / 6371.2}
    cs_car = gsl_cs.GEOCCartesian(coords=coords_in, ut=dts)
    cs_sph = cs_car.to_spherical()
    orbits = {
        'GEO_LAT': cs_sph['lat'],
        'GEO_LON': cs_sph['lon'],
        'GEO_ALT': cs_sph['height'],
        'GEO_X': coords['X'],
        'GEO_Y': coords['Y'],
        'GEO_Z': coords['Z'],
        'GSE_X': coords_gse['X'],
        'GSE_Y': coords_gse['Y'],
        'GSE_Z': coords_gse['Z'],
        'DATETIME': dts,
    }

    if to_AACGM:
        cs_aacgm = cs_sph.to_AACGM(append_mlt=True)
        orbits.update(
            **{
                'AACGM_LAT': cs_aacgm['lat'],
                'AACGM_LON': cs_aacgm['lon'],
                'AACGM_MLT': cs_aacgm['mlt'],
            }
        )
    return orbits


def check_option_sat_time():
    sat_ids = ['f11', 'f12', 'f13', 'f14', 'f15', 'f16', 'f17', 'f18', 'f19']
    # sat_ids = ['f15']

    dt_fr = datetime.datetime(2000, 1, 1, 0)
    dt_to = datetime.datetime(2021, 12, 31, 12, 59, 59)

    diff_months = (dt_to.year - dt_fr.year) * 12 + (dt_to.month - dt_fr.month) % 12

    for nm in range(diff_months + 1):
        thismonth = dttool.get_next_n_months(dt_fr, nm)
        thismonthend = dttool.get_last_day_of_month(thismonth, end=True)

        for sat_id in sat_ids:
            orbits = dmsp_orbits(dt_fr=thismonth, dt_to=thismonthend, sat_id=sat_id, to_AACGM='on')
            if orbits is None:
                continue
            fp = pathlib.Path(__file__).resolve().parent.resolve() / f'SSCWS_orbits_DMSP-{sat_id.upper()}_{thismonth.strftime("%Y%m")}_1min.nc'
            fp.parent.resolve().mkdir(parents=True, exist_ok=True)
            fnc = nc.Dataset(fp, 'w')
            fnc.createDimension('UNIX_TIME', orbits['DATETIME'].shape[0])

            fnc.title = f"DMSP-{sat_id.upper()} orbits from {thismonth.strftime('%Y%m%dT%H%M%S')} to {thismonthend.strftime('%Y%m%dT%H%M%S')}"
            time = fnc.createVariable('UNIX_TIME', np.float64, ('UNIX_TIME',))
            time.units = 'seconds since 1970-01-01 00:00:00.0'
            geo_lat = fnc.createVariable('GEO_LAT', np.float32, ('UNIX_TIME',))
            geo_lon = fnc.createVariable('GEO_LON', np.float32, ('UNIX_TIME',))
            geo_alt = fnc.createVariable('GEO_ALT', np.float32, ('UNIX_TIME',))
            geo_x = fnc.createVariable('GEO_X', np.float32, ('UNIX_TIME',))
            geo_y = fnc.createVariable('GEO_Y', np.float32, ('UNIX_TIME',))
            geo_z = fnc.createVariable('GEO_Z', np.float32, ('UNIX_TIME',))
            gse_x = fnc.createVariable('GSE_X', np.float32, ('UNIX_TIME',))
            gse_y = fnc.createVariable('GSE_Y', np.float32, ('UNIX_TIME',))
            gse_z = fnc.createVariable('GSE_Z', np.float32, ('UNIX_TIME',))
            aa_lat = fnc.createVariable('AACGM_LAT', np.float32, ('UNIX_TIME',))
            aa_lon = fnc.createVariable('AACGM_LON', np.float32, ('UNIX_TIME',))
            aa_mlt = fnc.createVariable('AACGM_MLT', np.float32, ('UNIX_TIME',))

            time_array = np.array(cftime.date2num(orbits['DATETIME'].flatten(), units='seconds since 1970-01-01 00:00:00.0'))
            time[::] = time_array[::]
            geo_lat[::] = orbits['GEO_LAT'][::]
            geo_lon[::] = orbits['GEO_LON'][::]
            geo_alt[::] = orbits['GEO_ALT'][::]
            geo_x[::] = orbits['GEO_X'][::]
            geo_y[::] = orbits['GEO_Y'][::]
            geo_z[::] = orbits['GEO_Z'][::]
            gse_x[::] = orbits['GSE_X'][::]
            gse_y[::] = orbits['GSE_Y'][::]
            gse_z[::] = orbits['GSE_Z'][::]
            aa_lat[::] = orbits['AACGM_LAT'][::]
            aa_lon[::] = orbits['AACGM_LON'][::]
            aa_mlt[::] = orbits['AACGM_MLT'][::]

            print(f"DMSP-{sat_id.upper()} orbits from {thismonth.strftime('%Y%m%dT%H%M%S')} to {thismonthend.strftime('%Y%m%dT%H%M%S')}")
            fnc.close()


def test():
    sat_id = 'f18'
    dt_fr = datetime.datetime(2010, 1, 1, 0)
    dt_to = datetime.datetime(2022, 1, 1, 12, 59, 59)
    orbits = dmsp_orbits(dt_fr=dt_fr, dt_to=dt_to, sat_id=sat_id)

    import matplotlib.pyplot as plt
    plt.plot(orbits['DATETIME'], orbits['GEO_LAT'])


if __name__ == "__main__":
    check_option_sat_time()



# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"


import h5py
import numpy as np
import cftime
import re
import datetime

import geospacelab.toolbox.utilities.pylogging as mylog
import geospacelab.toolbox.utilities.pydatetime as dttool

pulse_code_dict = {
    'alternating': 97,
    'barker': 98,
    'single pulse': 115,
}

antenna_code_dict = {
    'zenith': 32,
    'misa': 31,
}

var_name_dict = {
    'AZ1': 'az1',
    'AZ2': 'az2',
    'EL1': 'el1',
    'EL2': 'el2',
    'PULSE_LENGTH': 'pl',
    'T_SYS': 'systmp',
    'POWER_NORM': 'pnrmd',
    'P_Tx': 'power',
    'MODE_TYPE': 'mdtyp',
    'POWER_LENGTH_F': 'pulf',
    'LAG_SPACING': 'dtau',
    'IPP': 'ipp',
    'f_Tx': 'tfreq',
    'v_PHASE_Tx': 'vtx',
    'v_PHASE_Tx_err': 'dvtx',
    'SCAN_TYPE': 'scntyp',
    'CYCN': 'cycn',
    'POSN': 'posn',
    'RANGE_RES': 'mresl',
    'RANGE': 'range',
    'SNR': 'snp3',
    'RESIDUAL': 'wchsq',
    'STATUS': 'gfit',
    'FIT_TYPE': 'fit_type',
    'FPI_QUALITY': 'fpi_dataqual',
    'ACF_NORM': 'fa',
    'ACF_NORM_ERR': 'dfa',
    'n_pp': 'popl',
    'n_pp_err': 'dpopl',
    'n_e': 'ne',
    'n_e_err': 'dne',
    'T_i': 'ti',
    'T_i_err': 'dti',
    'T_r': 'tr',
    'T_r_err': 'dtr',
    'nu_i': 'co',
    'nu_i_err': 'dco',
    'v_i_los': 'vo',
    'v_i_los_err': 'dvo',
    'comp_H_p': 'ph+',
    'comp_H_p_err': 'dph+',
    'comp_mix': 'pm',
    'comp_mix_err': 'dpm',
    'v_DOP_los': 'vdopp',
    'v_DOP_los_err': 'dvdopp',
    'HEIGHT': 'gdalt'
}


class Loader:
    """
    :param file_path: the file's full path
    :type file_path: pathlib.Path object
    :param file_type: the specific file type for the file being loaded. Options: ['TEC-MAT'], 'TEC-LOS', 'TEC-sites')
    :type file_type: str
    :param load_data: True, load without calling the method "load_data" separately.
    :type load_data: bool
    """
    def __init__(self, file_path, antenna='', pulse_code='', pulse_length=0, load_data=True):
        self.file_path = file_path
        self.antenna = antenna
        self.pulse_code = pulse_code
        self.pulse_length = pulse_length
        self.variables = {}
        self.metadata = {}

        self.done = False
        if load_data:
            self.load()

    def load(self):
        variables = {}
        metadata = {}
        with h5py.File(self.file_path, 'r') as fh5:
            data_fh5 = fh5['Data']
            array_layouts = list(data_fh5['Array Layout'].keys())

            antenna_codes = []
            pulse_codes = []
            pulse_lengths = []
            match = 0
            ind_array = None
            for i, array_layout_str in enumerate(array_layouts):
                res = re.search(r'kinst=([\d.]+).*mdtyp=([\d.]+).*pl=([\d.]+)', array_layout_str)
                antenna_code = int(float(res.groups()[0]))
                pulse_code = int(float(res.groups()[1]))
                pulse_length = float(res.groups()[2]) * 1e6
                antenna_codes.append(antenna_code)
                pulse_codes.append(pulse_code)
                pulse_lengths.append(pulse_length)
                if not str(self.antenna) or not str(self.pulse_code) or self.pulse_length == 0:
                    continue
                if antenna_code != antenna_code_dict[self.antenna] or \
                        pulse_code != pulse_code_dict[self.pulse_code] or \
                        pulse_length != self.pulse_length:
                    continue
                match = match + 1
                ind_array = i
            antenna_code_dict_r = {v: k for k, v in antenna_code_dict.items()}
            pulse_code_dict_r = {v: k for k, v in pulse_code_dict.items()}
            if match != 1:
                mylog.StreamLogger.error("The inputs do not match the Array Layouts! Check the Array Layout info below:")
                mylog.simpleinfo.info(
                        "{:10s}{:20s}{:30s}{:30s}".format(
                            'No.',
                            'antenna',
                            'pulse_code',
                            'pulse_length',
                        )
                    )
                for ind, (antenna_code, pulse_code, pulse_length) \
                        in enumerate(zip(antenna_codes, pulse_codes, pulse_lengths)):

                    mylog.simpleinfo.info(
                        "{:<10d}{:20s}{:30s}{:<30f}".format(
                            ind+1,
                            antenna_code_dict_r[antenna_code],
                            pulse_code_dict_r[pulse_code],
                            pulse_length,
                        )
                    )
                raise ValueError("Check the message above!")

            metadata['antenna'] = antenna_code_dict_r[antenna_codes[ind_array]]
            metadata['pulse_code'] = pulse_code_dict_r[pulse_codes[ind_array]]
            metadata['pulse_length'] = pulse_lengths[ind_array]

            vars_fh5 = {}
            array_layout_str = array_layouts[ind_array]
            fh5_vars_1d = data_fh5['Array Layout'][array_layout_str]['1D Parameters']
            for var_name in fh5_vars_1d.keys():
                if var_name == 'Data Parameters':
                    continue
                vars_fh5[var_name] = np.array(fh5_vars_1d[var_name])[:, np.newaxis]
            fh5_vars_2d = data_fh5['Array Layout'][array_layout_str]['2D Parameters']
            for var_name in fh5_vars_2d.keys():
                if var_name == 'Data Parameters':
                    continue
                vars_fh5[var_name] = np.array(fh5_vars_2d[var_name]).T
            vars_fh5['range'] = np.array(data_fh5['Array Layout'][array_layout_str]['range'])[np.newaxis, :]
            vars_fh5['timestamps'] = np.array(data_fh5['Array Layout'][array_layout_str]['timestamps'])[:, np.newaxis]
            for var_name, var_name_fh5 in var_name_dict.items():
                if var_name_fh5 not in vars_fh5.keys():
                    mylog.StreamLogger.warning(f"The requested variable {var_name_fh5} does not exist in the data file!")
                    variables[var_name] = None
                    continue
                variables[var_name] = vars_fh5[var_name_fh5]

            variables['comp_O_p'] = 1. - variables['comp_mix'] - variables['comp_H_p']
            variables['comp_O_p_err'] = np.sqrt(variables['comp_mix_err']**2 + variables['comp_H_p_err']**2)

            # need to be check when AZ close to 0.
            variables['AZ'] = variables['AZ1'] % 360.
            variables['EL'] = variables['EL1']

            variables['RANGE'] = np.tile(vars_fh5['range'], [variables['n_e'].shape[0], 1])
            variables['DATETIME'] = dttool.convert_unix_time_to_datetime_cftime(vars_fh5['timestamps'])
            variables['T_e'] = variables['T_i'] * variables['T_r']
            variables['T_e_err'] = variables['T_e'] * np.sqrt((variables['T_i_err'] / variables['T_i']) ** 2
                                                              + (variables['T_r_err'] / variables['T_r']) ** 2)

        self.variables = variables
        self.metadata = metadata


if __name__ == "__main__":
    import pathlib
    fp = pathlib.Path("/home/lei/afys-data/Madrigal/Millstone_ISR/2016/20160314/Millstone_ISR_combined_20160314.005.hdf5")
    Loader(file_path=fp)